from collections.abc import Mapping, MutableMapping
from ctypes import ArgumentError
from pathlib import Path
from typing import Any

import numpy as np
from zarr.storage import KVStore, Store, _path_to_prefix, attrs_key, init_array, init_group
from zarr.util import json_dumps, normalize_shape, normalize_storage_path

from ._wsi_reader import ReaderBase


def init_attrs(store: MutableMapping, attrs: Mapping[str, Any], path: str | None = None):
    path = normalize_storage_path(path)
    path = _path_to_prefix(path)
    store[path + attrs_key] = json_dumps(attrs)


def create_meta_store(reader: ReaderBase, tilesize: int) -> dict[str, bytes]:
    """Creates a dict containing the zarr metadata for the multiscale openslide image."""
    store = {}
    root_attrs = {
        "multiscales": [
            {
                "name": Path(reader.path).name,
                "datasets": [{"path": str(i)} for i in range(reader.level_count)],
                "version": "0.1",
            }
        ],
        "metadata": dict(reader.properties),
    }
    init_group(store)
    init_attrs(store, root_attrs)

    for i, (x, y) in enumerate(reader.level_dimensions):
        init_array(
            store,
            path=str(i),
            shape=normalize_shape((y, x, 3)),
            chunks=(tilesize, tilesize, 3),
            fill_value=0,
            dtype="|u1",
            compressor=None,
        )
        suffix = str(i) if i != 0 else ""
        init_attrs(store, {"_ARRAY_DIMENSIONS": [f"Y{suffix}", f"X{suffix}", "S"]}, path=str(i))
    return store


def _parse_chunk_path(path: str):
    """Returns x,y chunk coords and pyramid level from string key"""
    level, ckey = path.split("/")
    y, x, _ = map(int, ckey.split("."))
    return x, y, int(level)


class WsiStore(Store):
    """Wraps an OpenSlide object as a multiscale Zarr Store.

    Parameters
    ----------
    path: str
        The file to open with OpenSlide.
    tilesize: int
        Desired "chunk" size for zarr store (default: 512).
    """

    def __init__(self, reader: ReaderBase, tilesize: int = 512):
        self._reader = reader
        self._tilesize = tilesize
        self._store = create_meta_store(self._reader, tilesize)
        self._writeable = False
        self._erasable = False

    def __contains__(self, key: str):
        if key in self._store:
            return True
        else:
            try:
                _, _, level = _parse_chunk_path(key)
            except ValueError:
                return False
            else:
                return f"{level}/.zarray" in self._store

    def __getitem__(self, key: str):
        if key in self._store:
            # key is for metadata
            return self._store[key]

        # key should now be a path to an array chunk
        # e.g '3/4.5.0' -> '<level>/<chunk_key>'
        try:
            x, y, level = _parse_chunk_path(key)
            location = self._ref_pos(x, y, level)
            size = (self._tilesize, self._tilesize)
            tile = self._reader.read_region(location, level, size)
        except ArgumentError as err:
            # Can occur if trying to read a closed slide
            raise ArgumentError(err)
        except Exception:
            # TODO: probably need better error handling.
            # If anything goes wrong, we just signal the chunk
            # is missing from the store.
            raise KeyError(key)
        return np.array(tile)  # .tobytes()

    def __eq__(self, other):
        return isinstance(other, WsiStore) and self._reader.path == other._reader.path

    def __setitem__(self, key, val):
        raise PermissionError("ZarrStore is read-only")

    def __delitem__(self, key):
        raise PermissionError("ZarrStore is read-only")

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return sum(1 for _ in self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _ref_pos(self, x: int, y: int, level: int):
        dsample = self._reader.level_downsamples[level]
        xref = int(x * dsample * self._tilesize)
        yref = int(y * dsample * self._tilesize)
        return xref, yref

    def keys(self):
        return self._store.keys()

    def close(self):
        self._reader.close()

    # Retrieved from napari-lazy-openslide PR#16
    def __getstate__(self):
        return {"_path": self._path, "_tilesize": self._tilesize}

    def __setstate__(self, newstate):
        path = newstate["_path"]
        tilesize = newstate["_tilesize"]
        self.__init__(path, tilesize)

    def rename(self):
        raise PermissionError(f'{type(self)} is not erasable, cannot call "rename"')

    def rmdir(self, path: str = "") -> None:
        raise PermissionError(f'{type(self)} is not erasable, cannot call "rmdir"')

    @property
    def store(self) -> KVStore:
        return KVStore(self)



from .images import write_image, align
from .points import write_transcripts
from .table import write_cell_categories, write_gene_counts, save_column_csv
from .shapes import write_polygons
from .converter import write, write_metadata
from .utils import str_cell_id, int_cell_id, add_explorer_selection



import logging
from math import ceil, floor

import geopandas as gpd
import numpy as np
import shapely
import shapely.affinity
import skimage
from geopandas import GeoDataFrame
from shapely.errors import GEOSException
from shapely.geometry import LinearRing, MultiPolygon, Polygon
from skimage.measure._regionprops import RegionProperties

from .. import settings
from .validate import ensure_polygon

log = logging.getLogger(__name__)


def vectorize(mask: np.ndarray, tolerance: float | None = None, smooth_radius_ratio: float = 0.1) -> gpd.GeoDataFrame:
    """Convert a cells mask to multiple `shapely` geometries. Each shape is smoothen and converted to a single polygon. Inspired from [VPT](https://github.com/Vizgen/vizgen-postprocessing).

    Args:
        mask: A cell mask. Non-null values correspond to cell ids
        tolerance: Tolerance parameter used by `shapely` during simplification. By default, define the tolerance automatically.
        smooth_radius_ratio: Ratio of the cell radius used to smooth the cell polygon.

    Returns:
        GeoDataFrame of polygons representing each cell ID of the mask
    """
    max_cells = mask.max()

    if max_cells == 0:
        log.warning("No cell was returned by the segmentation")
        return gpd.GeoDataFrame(geometry=[])

    cells = _vectorize_mask(mask)

    mean_radius = np.sqrt(cells.area / np.pi).mean()
    smooth_radius = mean_radius * smooth_radius_ratio

    tolerance = _default_tolerance(mean_radius) if tolerance is None else tolerance

    cells.geometry = cells.geometry.map(lambda cell: _smoothen_cell(cell, smooth_radius, tolerance))
    cells = cells[~cells.is_empty]

    return cells


def _region_props_to_multipolygon(region_props: RegionProperties, allow_holes: bool) -> Polygon | MultiPolygon:
    mask = np.pad(region_props.image, 1)
    contours = skimage.measure.find_contours(mask, 0.5)

    rings = [LinearRing(contour[:, [1, 0]]) for contour in contours if contour.shape[0] >= 4]

    exteriors = [ring for ring in rings if ring.is_ccw]

    if allow_holes:
        holes = [ring for ring in rings if not ring.is_ccw]

    def _to_polygon(exterior: LinearRing) -> Polygon:
        exterior_poly = Polygon(exterior)

        _holes = None
        if allow_holes:
            _holes = [hole.coords for hole in holes if exterior_poly.contains(Polygon(hole))]

        return Polygon(exterior, holes=_holes)

    polygon = [_to_polygon(exterior) for exterior in exteriors]
    polygon = MultiPolygon(polygon) if len(polygon) > 1 else polygon[0]

    yoff, xoff, *_ = region_props.bbox
    return shapely.affinity.translate(polygon, xoff - 1, yoff - 1)  # remove padding offset


def _vectorize_mask(mask: np.ndarray, allow_holes: bool = False) -> GeoDataFrame:
    if mask.max() == 0:
        return GeoDataFrame(geometry=[])

    regions = skimage.measure.regionprops(mask)

    return GeoDataFrame(geometry=[_region_props_to_multipolygon(region, allow_holes) for region in regions])


def pixel_outer_bounds(bounds: tuple[int, int, int, int]) -> tuple[int, int, int, int]:
    return [floor(bounds[0]), floor(bounds[1]), ceil(bounds[2]) + 1, ceil(bounds[3]) + 1]


def _smoothen_cell(cell: MultiPolygon, smooth_radius: float, tolerance: float) -> Polygon:
    """Smoothen a cell polygon

    Args:
        cell_id: MultiPolygon representing a cell
        smooth_radius: radius used to smooth the cell polygon
        tolerance: tolerance used to simplify the cell polygon

    Returns:
        Shapely polygon representing the cell, or an empty Polygon if the cell was empty after smoothing
    """
    try:
        if smooth_radius > 0:
            cell = cell.buffer(-smooth_radius).buffer(2 * smooth_radius).buffer(-smooth_radius)
        if tolerance > 0:
            cell = cell.simplify(tolerance)
    except GEOSException:
        log.warning(f"Failed to smoothen cell with {smooth_radius=} and tolerance {tolerance=}.")
        return Polygon()

    return ensure_polygon(cell)


def _default_tolerance(mean_radius: float) -> float:
    return settings.simplification_tolerance or min(mean_radius * 0.03, 1)



from .bins import aggregate_bins
from .channels import aggregate_channels
from .transcripts import count_transcripts
from .aggregation import aggregate, Aggregator
from .overlay import overlay_segmentation



from __future__ import annotations

import base64
from io import BytesIO

import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.figure import Figure

from .css import BULMA_CSS


class Renderable:
    """Object that can be transformed to string representing HTML"""

    _children: list[Renderable] | Renderable | None = None

    @property
    def children(self) -> list[Renderable]:
        if hasattr(self, "_children") and self._children is not None:
            if isinstance(self._children, list):
                return self._children
            return [self._children]
        return []

    @property
    def children_html(self) -> str:
        return "".join([str(child) for child in self.children])

    def children_rec(self) -> list[Renderable]:
        return [self] + [cc for child in self.children for cc in child.children_rec()]

    def __str__(self) -> str:
        return self.children_html


class Title(Renderable):
    """Report title"""

    def __init__(self, text: str, level: int, subtitle: bool = False) -> None:
        self.text = text
        self.level = level
        self.subtitle = subtitle

    def __str__(self) -> str:
        return f"""<h1 class="{"subtitle" if self.subtitle else "title"} is-{self.level}">{self.text}</h1>"""


class Paragraph(Renderable):
    """Report paragraph"""

    def __init__(self, text: str) -> None:
        self.text = text

    def __str__(self) -> str:
        return f"""<p>{self.text}</p>"""


class Message(Renderable):
    """Colored message"""

    def __init__(self, text: str, is_light: bool = True, color: str = "primary") -> None:
        self.text = text
        self.color = color
        self.is_light = is_light

    def __str__(self) -> str:
        return f"""
        <div class='notification is-{self.color} {"is-light" if self.is_light else ""}'>
            {self.text}
        </div>
        """


class Block(Renderable):
    """Block, i.e. padded div"""

    def __init__(self, content: list[Renderable]) -> None:
        self._children = content

    def __str__(self) -> str:
        return f"""<div class="block">{self.children_html}</div>"""


class CodeBlock(Renderable):
    """Block of code, like in the terminal"""

    def __init__(self, text: str) -> None:
        self.text = text

    def __str__(self) -> str:
        return f"""<pre>{self.text}</pre>"""


class ProgressBar(Renderable):
    """Progress bar"""

    def __init__(
        self,
        value: float,
        valuemax: int = 1,
        text: str | None = None,
        color: str = "primary",
        is_light: bool = False,
    ) -> None:
        self.value = value
        self.valuemax = valuemax
        self.text = text
        self.color = color
        self.is_light = is_light

    def __str__(self) -> str:
        return f"""
        {"" if self.text is None else Paragraph(self.text)}
        <progress class="progress is-{self.color} {"is-light" if self.is_light else ""}"
            value="{self.value}"
            max="{self.valuemax}">{self.value}
        </progress>
    """


class Section(Renderable):
    """Section of the report"""

    def __init__(self, name: str, content: list[Section] | None = None) -> None:
        self.name = name
        self._children = content
        self.subtitle = False

    @property
    def id(self) -> str:
        return self.name.lower().replace(" ", "-")

    def __str__(self) -> str:
        return f"""
        <article class="message is-dark" id="{self.id}">
            <div class="message-header">
                <p>{self.name}</p>
            </div>
            <div class="message-body">
                {self.children_html}
            </div>
        </article>
        """


class SubSection(Section):
    """Sub-section of the report"""

    def __init__(self, *args: int, **kwargs: int):
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        return f"""
        <section id="{self.id}" class="py-2">
            {Title(self.name, 4, subtitle=True)}
            {self.children_html}
        </section>
        """


class NavbarItem(Renderable):
    """One item in the nav bar"""

    def __init__(self, section: Section) -> None:
        self.section = section

    def subsections(self) -> str:
        li = [f"<li><a href='#{subsection.id}'>{subsection.name}</a></li>" for subsection in self.section.children]
        return "".join(li)

    def __str__(self) -> str:
        return f"""
    <p class="menu-label">{self.section.name}</p>
    <ul class="menu-list">
        {self.subsections()}
    </ul>
    """


class Navbar(Renderable):
    """Left nav bar"""

    def __init__(self, sections: list[Section]) -> None:
        self._children = [NavbarItem(section) for section in sections]

    def __str__(self) -> str:
        return f"""
        {Title("Sopa report", 3)}
        {Message("This report was generated <br />by the <a href='https://github.com/gustaveroussy/sopa'>Sopa</a> HTML engine.")}
        {self.children_html}
    """


class Columns(Renderable):
    """Flex columns containers"""

    def __init__(self, content: list[Renderable], is_centered: bool = True) -> None:
        self._children = content
        self.is_centered = is_centered

    def __str__(self) -> str:
        return f"""
    <div block style="display: flex; {"justify-content: center;" if self.is_centered else ""}">
        {self.children_html}
    </div>
    """


class Image(Renderable):
    """Image renderer"""

    def __init__(self, fig: Figure, width: float = 50, extension: str = "png", pretty_legend: bool = True):
        self.fig = fig
        self.width = width
        self.extension = extension
        self.pretty_legend = pretty_legend

    def make_figure_pretty(self) -> None:
        if self.pretty_legend and _has_handles(self.fig):
            self.fig.legend(bbox_to_anchor=(1.04, 0.5), loc="center left", borderaxespad=0, frameon=False)
        sns.despine(fig=self.fig, offset=10, trim=True)

    def encod(self) -> str:
        self.make_figure_pretty()
        tmpfile = BytesIO()
        self.fig.savefig(tmpfile, format=self.extension, transparent=True, bbox_inches="tight")
        plt.close()
        return base64.b64encode(tmpfile.getvalue()).decode("utf-8")

    def __str__(self) -> str:
        return (
            f"""<img src=\'data:image/{self.extension};base64,{self.encod()}\'  width="{self.width}%" height="auto"/>"""
        )


def _has_handles(fig: Figure) -> bool:
    return any(len(ax.get_legend_handles_labels()[0]) for ax in fig.get_axes())


class Root(Renderable):
    """Whole report generator"""

    def __init__(self, sections: list[Section], doc_title: str = "Sopa report"):
        self.doc_title = doc_title
        self._children = sections
        self.nav = Navbar(sections)

    def sanity_check(self) -> None:
        section_ids = [section.id for section in self.children]
        assert len(section_ids) == len(set(section_ids)), "Sections IDs must be unique"

        subsections_ids = [sub.id for section in self.children for sub in section.children]
        assert len(subsections_ids) == len(set(subsections_ids)), "Subsections IDs must be unique"

    def write(self, path: str) -> None:
        self.sanity_check()

        with open(path, "w", encoding="utf-8") as f:
            f.write(str(self))

    def __str__(self) -> str:
        return f"""
    <!DOCTYPE html>
    <html>
    <head>
        <meta charset="utf-8" />
        <meta name="viewport" content="width=device-width, initial-scale=1" />
        <title>{self.doc_title}</title>
        <style>
        {BULMA_CSS}
        .menu {{
            position: sticky;
            flex: 0 0 260px;
            overflow-y: auto;
            height: 100vh;
            top: 0;
        }}
        </style>
    </head>
    <body>
        <div class="is-flex is-flex-direction-row">
            <div class="mt-5 ml-5 menu">
                {self.nav}
            </div>
            <div class="p-5 block" style="flex: 1; overflow: hidden">
                {self.children_html}
            </div>
        </div>
    </body>
    </html>
    """



import logging
from pathlib import Path

import xarray as xr
from spatialdata import SpatialData
from spatialdata.models import Image2DModel

from sopa.constants import SopaAttrs
from sopa.io.reader.utils import _default_image_kwargs, _image_int_dtype

log = logging.getLogger(__name__)


def aicsimageio(
    path: Path,
    z_stack: int = 0,
    image_models_kwargs: dict | None = None,
    aics_kwargs: dict | None = None,
) -> SpatialData:
    """Read an image using [AICSImageIO](https://github.com/AllenCellModeling/aicsimageio). It supports special formats such as `ND2`, `CZI`, `LIF`, or `DV`.

    !!! note "Extra dependencies"
        To use this reader, you'll need the `aicsimageio` dependency (`pip install aicsimageio`). To read `.czi` images, you'll also need to install `aicspylibczi` (for instance `pip install aicspylibczi`).

    Args:
        path: Path to the image file
        z_stack: (Only for 3D images) Index of the stack in the z-axis to use.
        image_models_kwargs: Keyword arguments passed to `spatialdata.models.Image2DModel`.
        aics_kwargs: Keyword arguments passed to `aicsimageio.AICSImage`.

    Returns:
        A `SpatialData` object with a 2D-image of shape `(C, Y, X)`
    """
    log.warning("The `aicsimageio` reader is deprecated. Use `sopa.io.bioio` instead.")
    image_models_kwargs, _ = _default_image_kwargs(image_models_kwargs, None)
    aics_kwargs = {} if aics_kwargs is None else aics_kwargs

    try:
        from aicsimageio import AICSImage
    except ImportError:
        raise ImportError("You need to install aicsimageio, e.g. by running `pip install aicsimageio`")

    xarr: xr.DataArray = AICSImage(path, **aics_kwargs).xarray_dask_data

    assert len(xarr.coords["T"]) == 1, f"Only one time dimension is supported, found {len(xarr.coords['T'])}."

    if len(xarr.coords["Z"]) > 1:
        log.info(f"3D image found, only reading {z_stack:=}")

    xarr = xarr.isel(T=0, Z=z_stack).rename({"C": "c", "Y": "y", "X": "x"})
    xarr = _image_int_dtype(xarr)

    image = Image2DModel.parse(xarr, c_coords=xarr.coords["c"].values, **image_models_kwargs)

    return SpatialData(images={"image": image}, attrs={SopaAttrs.CELL_SEGMENTATION: "image"})


def bioio(
    path: Path,
    z_stack: int = 0,
    image_models_kwargs: dict | None = None,
    bioio_kwargs: dict | None = None,
) -> SpatialData:
    """Read an image using [bioio](https://github.com/bioio-devs/bioio). It supports special formats such as `ND2`, `CZI`, `LIF`, or `DV`.

    !!! note "Extra dependencies"
        To use this reader, you'll need the `bioio` dependency (`pip install bioio`). You may need extra dependencies specific to your format, see their [documentation](https://bioio-devs.github.io/bioio/OVERVIEW.html#reader-installation).

    Args:
        path: Path to the image file
        z_stack: (Only for 3D images) Index of the stack in the z-axis to use.
        image_models_kwargs: Keyword arguments passed to `spatialdata.models.Image2DModel`.
        bioio_kwargs: Keyword arguments passed to `bioio.BioImage`.

    Returns:
        A `SpatialData` object with a 2D-image of shape `(C, Y, X)`
    """
    image_models_kwargs, _ = _default_image_kwargs(image_models_kwargs, None)
    bioio_kwargs = {} if bioio_kwargs is None else bioio_kwargs

    try:
        from bioio import BioImage
    except ImportError:
        raise ImportError("You need to install bioio, e.g. by running `pip install bioio`")

    xarr: xr.DataArray = BioImage(path, **bioio_kwargs).xarray_dask_data

    assert len(xarr.coords["T"]) == 1, f"Only one time dimension is supported, found {len(xarr.coords['T'])}."

    if len(xarr.coords["Z"]) > 1:
        log.info(f"3D image found, only reading {z_stack:=}")

    xarr = xarr.isel(T=0, Z=z_stack).rename({"C": "c", "Y": "y", "X": "x"})
    xarr = _image_int_dtype(xarr)

    image = Image2DModel.parse(xarr, c_coords=xarr.coords["c"].values, **image_models_kwargs)

    return SpatialData(images={"image": image}, attrs={SopaAttrs.CELL_SEGMENTATION: "image"})



import ast
import logging

import typer

from .annotate import app_annotate
from .explorer import app_explorer
from .patchify import app_patchify
from .resolve import app_resolve
from .segmentation import app_segmentation
from .utils import SDATA_HELPER

log = logging.getLogger(__name__)

app = typer.Typer()
app.add_typer(
    app_annotate,
    name="annotate",
    help="Perform cell-type annotation (based on transcripts and/or channel intensities)",
)
app.add_typer(
    app_explorer,
    name="explorer",
    help="Convertion to the Xenium Explorer's inputs, and image alignment",
)
app.add_typer(
    app_segmentation,
    name="segmentation",
    help="Perform cell segmentation on patches. NB: for `baysor`, use directly the `baysor` command line.",
)
app.add_typer(app_resolve, name="resolve", help="Resolve the segmentation conflicts over patches overlaps")
app.add_typer(
    app_patchify,
    name="patchify",
    help="Create patches with overlaps. Afterwards, segmentation will be run on each patch",
)


@app.command()
def convert(
    data_path: str = typer.Argument(
        help="Path to one data sample (most of the time, this corresponds to a directory with images files and eventually a transcript file)"
    ),
    technology: str = typer.Option(
        None,
        help="Name of the technology used to collected the data (`xenium`/`merfish`/`cosmx`/`phenocycler`/`macsima`/`hyperion`)",
    ),
    sdata_path: str = typer.Option(
        None,
        help="Optional path to write the SpatialData object. If not provided, will write to the `{data_path}.zarr` directory",
    ),
    config_path: str = typer.Option(
        None,
        help="Path to the snakemake config. This can be useful in order not to provide the `--technology` and the `--kwargs` arguments",
    ),
    overwrite: bool = typer.Option(
        False, help="Whether to overwrite the existing SpatialData object if already existing"
    ),
    kwargs: str = typer.Option(
        default={},
        callback=ast.literal_eval,
        help="Dictionary provided to the reader function as kwargs",
    ),
):
    """Read any technology data as a SpatialData object and save it as a `.zarr` directory.

    Either `--technology` or `--config-path` has to be provided."""
    import shutil
    from pathlib import Path

    from spatialdata import SpatialData

    from sopa import io
    from sopa.constants import SopaFiles, SopaKeys

    sdata_path: Path = Path(data_path).with_suffix(".zarr") if sdata_path is None else Path(sdata_path)

    if sdata_path.exists():
        if overwrite:
            cache_dir = sdata_path.resolve() / SopaFiles.SOPA_CACHE_DIR
            if cache_dir.exists() and cache_dir.is_dir():
                shutil.rmtree(cache_dir)
            if not any(sdata_path.iterdir()):
                sdata_path.rmdir()  # remove empty directory
        else:
            assert not any(sdata_path.iterdir()), (
                f"Zarr directory {sdata_path} already exists. Sopa will not continue to avoid overwritting files. Use the `--overwrite` flag to overwrite it."
            )
            sdata_path.rmdir()  # remove empty directory

    assert technology is not None or config_path is not None, "Provide the argument `--technology` or `--config-path`"

    if config_path is not None:
        assert not kwargs, "Provide either a path to a config, or some kwargs, but not both"
        with open(config_path) as f:
            import yaml

            config = yaml.safe_load(f)

        technology = config["read"]["technology"]
        kwargs = config["read"]["kwargs"]

    assert hasattr(io, technology), (
        f"Technology {technology} unknown. Currently available: xenium, merscope, visium_hd, cosmx, phenocycler, hyperion, macsima"
    )

    sdata: SpatialData = getattr(io, technology)(data_path, **kwargs)

    io.standardize.sanity_check(sdata)

    assert SopaKeys.TABLE not in sdata.tables, (
        f"sdata.tables['{SopaKeys.TABLE}'] exists. Delete it you want to use sopa, to avoid conflicts with future table generation"
    )

    log.info(f"Writing the following spatialdata object to {sdata_path}:\n{sdata}")

    sdata.write(sdata_path, overwrite=overwrite)


@app.command()
def aggregate(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    aggregate_genes: bool = typer.Option(None, help="Whether to aggregate the genes (counts) inside each cell"),
    aggregate_channels: bool = typer.Option(
        False, help="Whether to aggregate the channels (intensity) inside each cell"
    ),
    expand_radius_ratio: float = typer.Option(
        default=None,
        help="Cells polygons will be expanded by `expand_radius_ratio * mean_radius` for channels averaging **only**. This help better aggregate boundary stainings",
    ),
    min_transcripts: int = typer.Option(0, help="Cells with less transcript than this integer will be filtered"),
    min_intensity_ratio: float = typer.Option(
        0,
        help="Cells whose mean channel intensity is less than `min_intensity_ratio * quantile_90` will be filtered",
    ),
    image_key: str = typer.Option(
        None,
        help="Optional image key of the SpatialData object. By default, considers the only one image. It can be useful if another image is added later on",
    ),
    gene_column: str = typer.Option(
        None, help="Optional column of the transcript dataframe to be used as gene names. Inferred by default."
    ),
    method_name: str = typer.Option(
        None, help="If segmentation was performed with a generic method, this is the name of the method used."
    ),
):
    """Create an `anndata` table containing the transcript count and/or the channel intensities per cell"""
    import sopa
    from sopa.io.standardize import read_zarr_standardized

    if gene_column:
        assert aggregate_genes is not False
        aggregate_genes = True

    sdata = read_zarr_standardized(sdata_path)

    sopa.aggregate(
        sdata,
        aggregate_genes=aggregate_genes,
        aggregate_channels=aggregate_channels,
        image_key=image_key,
        shapes_key=method_name,
        gene_column=gene_column,
        min_transcripts=min_transcripts,
        expand_radius_ratio=expand_radius_ratio,
        min_intensity_ratio=min_intensity_ratio,
    )


@app.command()
def scanpy_preprocess(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    table_key: str = typer.Option("table", help="Key of the table in the `SpatialData` object to be preprocessed"),
    resolution: float = typer.Option(1.0, help="Resolution parameter for the leiden clustering"),
    check_counts: bool = typer.Option(True, help="Whether to check that adata.X contains counts"),
    hvg: bool = typer.Option(
        False, help="Whether to compute highly variable genes before computing the UMAP and clustering"
    ),
):
    """Optional scanpy table preprocessing (log1p, UMAP, leiden clustering) after aggregation/annotation."""
    import scanpy as sc
    from anndata import AnnData

    import sopa
    from sopa.io.standardize import read_zarr_standardized

    sdata = read_zarr_standardized(sdata_path)

    adata: AnnData = sdata.tables[table_key]

    if check_counts:
        assert adata.X.max() > 10, (
            f"adata.X looks already log-normalized (max={adata.X.max()}). If you are sure it contains counts, use `--no-check-counts`"
        )

    adata.layers["counts"] = adata.X.copy()

    sc.pp.normalize_total(adata)
    sc.pp.log1p(adata)

    if hvg:
        sc.pp.highly_variable_genes(adata)

        if adata.var["highly_variable"].sum() <= 50:
            log.warning("Less than 50 HVG found. They will not be used for the UMAP and leiden clustering.")
            highly_variable = adata.var["highly_variable"]
            del adata.var["highly_variable"]  # else sc.pp.neighbors will crash

    sc.pp.neighbors(adata)
    sc.tl.umap(adata)
    sc.tl.leiden(adata, resolution=resolution, flavor="igraph")

    if hvg and "highly_variable" not in adata.var:
        adata.var["highly_variable"] = highly_variable  # put it back

    sopa.utils.add_spatial_element(sdata, table_key, adata)

    (sopa.utils.get_cache_dir(sdata) / "scanpy_preprocess").touch()


@app.command()
def report(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    path: str = typer.Argument(help="Path to the HTML report"),
    table_key: str = typer.Option(
        "table", help="Key of the table in the `SpatialData` object to be used for the report"
    ),
):
    """Create a HTML report of the pipeline run and some quality controls"""
    from sopa.io.report import write_report
    from sopa.io.standardize import read_zarr_standardized

    sdata = read_zarr_standardized(sdata_path)

    write_report(path, sdata, table_key=table_key)


def version_callback(value: bool):
    if value:
        from sopa import __version__

        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(None, "--version", callback=version_callback, help="Show the Sopa version and exit."),
):
    pass



import shutil
from typing import Callable

from spatialdata import SpatialData

from ...constants import SopaKeys
from ...utils import get_cache_dir, set_boundaries_attrs
from .. import StainingSegmentation, solve_conflicts


def custom_staining_based(
    sdata: SpatialData,
    method: Callable,
    channels: list[str] | str | None,
    image_key: str | None = None,
    min_area: float = 0,
    delete_cache: bool = True,
    recover: bool = False,
    clip_limit: float = 0.2,
    clahe_kernel_size: int | list[int] | None = None,
    gaussian_sigma: float = 1,
    cache_dir_name: str = SopaKeys.CUSTOM_BOUNDARIES,
    key_added: str = SopaKeys.CUSTOM_BOUNDARIES,
    min_patch_size: int = 10,
):
    """Run a generic staining-based segmentation model, and add a GeoDataFrame containing the cell boundaries.

    Args:
        sdata: A `SpatialData` object.
        method: A segmentation `callable` whose input is an image of shape `(C, Y, X)` and output is a cell mask of shape `(Y, X)`. Each mask value `>0` represent a unique cell ID. The `C` channels is determined by the `channels` argument.
        channels: Name of the channels to be used for segmentation (or list of channel names).
        image_key: Name of the image in `sdata` to be used for segmentation.
        min_area: Minimum area of a cell to be considered.
        delete_cache: Whether to delete the cache after segmentation.
        recover: If `True`, recover the cache from a failed segmentation, and continue.
        clip_limit: Parameter for skimage.exposure.equalize_adapthist (applied before running segmentation)
        clahe_kernel_size: Parameter for skimage.exposure.equalize_adapthist (applied before running segmentation)
        gaussian_sigma: Parameter for scipy gaussian_filter (applied before running segmentation)
        cache_dir_name: Name of the cache directory.
        key_added: Name of the key to be added to `sdata.shapes`.
        min_patch_size: Minimum patch size (in pixels) for both width and height. Patches smaller than this will be skipped to avoid segmentation errors.
    """
    temp_dir = get_cache_dir(sdata) / cache_dir_name

    segmentation = StainingSegmentation(
        sdata,
        method,
        channels,
        min_area=min_area,
        image_key=image_key,
        clip_limit=clip_limit,
        clahe_kernel_size=clahe_kernel_size,
        gaussian_sigma=gaussian_sigma,
        min_patch_size=min_patch_size,
    )
    segmentation.write_patches_cells(temp_dir, recover=recover)

    cells = StainingSegmentation.read_patches_cells(temp_dir)
    cells = solve_conflicts(cells)

    StainingSegmentation.add_shapes(sdata, cells, image_key=segmentation.image_key, key_added=key_added)

    set_boundaries_attrs(sdata, key_added)

    if delete_cache:
        shutil.rmtree(temp_dir)



import importlib.metadata
import logging
import sys
from ._logging import configure_logger

__version__ = importlib.metadata.version("sopa")

log = logging.getLogger("sopa")
configure_logger(log)

if not any(f"--{option}" in sys.argv for option in ["version", "help"]):  # no import for cli helpers
    from spatialdata import read_zarr  # will set `dataframe.query-planning` to False

    from ._settings import settings
    from . import utils
    from . import shapes
    from . import io
    from . import spatial
    from . import segmentation
    from .aggregation import aggregate, overlay_segmentation
    from .patches import make_transcript_patches, make_image_patches
    from .utils import get_spatial_image, get_spatial_element, to_intrinsic, get_boundaries



import logging
from pathlib import Path

import geopandas as gpd
import pandas as pd
from shapely import Polygon
from spatialdata import SpatialData
from spatialdata.models import ShapesModel

from ...constants import SopaAttrs
from ...utils import add_spatial_element, copy_transformations, get_spatial_element

log = logging.getLogger(__name__)


def explorer_file_path(path: str | Path, filename: str, is_dir: bool):
    path: Path = Path(path)

    if is_dir:
        path = path / filename

    return path


def is_valid_explorer_id(cell_id: str) -> bool:
    """Check if a cell ID is a valid Xenium Explorer ID"""
    if not hasattr(cell_id, "__len__"):
        return False
    if len(cell_id) == 10:
        return cell_id[:-2].isalpha() and cell_id[-2] == "-"
    if len(cell_id) == 8:
        return cell_id.isalpha()
    return False


def int_cell_id(explorer_cell_id: str | pd.Index) -> int | pd.Index:
    """Transforms an alphabetical cell id from the Xenium Explorer to an integer ID

    E.g., int_cell_id('aaaachba-1') = 10000

    Args:
        explorer_cell_id: An alphabetical cell ID or a pandas Index of many explorer cell IDs

    Returns:
        An integer or a pandas Index of integers representing cell IDs as indices"""
    if isinstance(explorer_cell_id, pd.Index):
        return explorer_cell_id.map(int_cell_id)

    assert isinstance(explorer_cell_id, str), "The cell ID must be a string or a pandas Index of strings"
    assert is_valid_explorer_id(explorer_cell_id), "The cell ID must be a valid Xenium Explorer ID"

    code = explorer_cell_id[:-2] if explorer_cell_id[-2] == "-" else explorer_cell_id
    coefs = [ord(c) - 97 for c in code][::-1]
    return sum(value * 16**i for i, value in enumerate(coefs))


def str_cell_id(cell_id: int | pd.Index) -> str | pd.Index:
    """Transforms an integer cell ID into an Xenium Explorer alphabetical cell id

    E.g., str_cell_id(10000) = 'aaaachba-1'

    Args:
        cell_id: An integer or a pandas Index of integers representing cell indices

    Returns:
        A string or a pandas Index of strings representing cell IDs in the Xenium Explorer format
    """
    if isinstance(cell_id, pd.Index):
        return cell_id.map(str_cell_id)

    assert isinstance(cell_id, int), "The cell ID must be an integer or a pandas Index of integers"

    coefs = []
    for _ in range(8):
        cell_id, coef = divmod(cell_id, 16)
        coefs.append(coef)
    return "".join([chr(97 + coef) for coef in coefs][::-1]) + "-1"


def _selection_to_polygon(df, pixel_size):
    return Polygon(df[["X", "Y"]].values / pixel_size)


def xenium_explorer_selection(path: str | Path, pixel_size: float = 0.2125, return_list: bool = False) -> Polygon:
    df = pd.read_csv(path, skiprows=2)

    if "Selection" not in df:
        polygon = _selection_to_polygon(df, pixel_size)
        return [polygon] if return_list else polygon

    return [_selection_to_polygon(sub_df, pixel_size) for _, sub_df in df.groupby("Selection")]


def add_explorer_selection(
    sdata: SpatialData,
    path: str,
    key_added: str = "explorer_selection",
    image_key: str | None = None,
    pixel_size: float = 0.2125,
):
    """After saving a selection on the Xenium Explorer, it will add all polygons inside `sdata.shapes[shapes_key]`

    Args:
        sdata: A `SpatialData` object
        path: The path to the `coordinates.csv` selection file
        key_added: The name to provide to the selection as shapes
        image_key: The original image name
        pixel_size: Number of microns in a pixel. It must be the same value as the one used in `sopa.io.write`
    """
    polys = xenium_explorer_selection(path, pixel_size=pixel_size, return_list=True)
    image = get_spatial_element(sdata.images, key=image_key or sdata.attrs.get(SopaAttrs.CELL_SEGMENTATION))

    geo_df = ShapesModel.parse(gpd.GeoDataFrame(geometry=polys), transformations=copy_transformations(image))
    add_spatial_element(sdata, key_added, geo_df)



import logging
from typing import Literal

from spatialdata import SpatialData

from ..constants import SopaAttrs, SopaKeys
from ..utils import get_spatial_element, get_spatial_image
from .patches import Patches2D
from .transcripts import OnDiskTranscriptPatches

log = logging.getLogger(__name__)


def make_image_patches(
    sdata: SpatialData,
    patch_width: int | None = 2000,
    patch_overlap: int = 50,
    image_key: str | None = None,
    roi_key: str | None = SopaKeys.ROI,
    key_added: str | None = None,
):
    """Create overlapping patches on an image. This can be used for image-based segmentation methods such as Cellpose, which will run on each patch.

    Args:
        sdata: A `SpatialData` object.
        patch_width: Width of the patches, in pixels. If `None`, creates only one patch.
        patch_overlap: Number of pixels of overlap between patches.
        image_key: Optional key of the image on which the patches will be made. If not provided, it is found automatically.
        roi_key: Optional name of the shapes that need to touch the patches. Patches that do not touch any shape will be ignored during segmentation. By default, uses `"region_of_interest"` if existing. If `None`, all patches will be used.
        key_added: Optional name of the patches to be saved. By default, uses `"image_patches"`.
    """
    image_key, _ = get_spatial_image(sdata, key=image_key, return_key=True)

    patches = Patches2D(
        sdata,
        image_key,
        patch_width=patch_width,
        patch_overlap=patch_overlap,
        roi_key=roi_key,
    )

    patches.add_shapes(key_added=key_added)


def make_transcript_patches(
    sdata: SpatialData,
    patch_width: float | int | None = 2000,
    patch_overlap: int = 50,
    points_key: str | None = None,
    prior_shapes_key: Literal["auto"] | str | None = None,
    unassigned_value: int | str | None = None,
    min_points_per_patch: int = 4000,
    write_cells_centroids: bool = False,
    roi_key: str | None = SopaKeys.ROI,
    key_added: str | None = None,
    **kwargs: int,
):
    """Create overlapping patches on a transcripts dataframe, and save it in a cache. This can be used for trancript-based segmentation methods such as Baysor or Proseg.

    !!! info "Prior segmentation usage"
        To save assign a prior segmentation to each transcript, you can use the `prior_shapes_key` argument:

        - If a segmentation has already been performed (for example an existing 10X-Genomics segmentation), use `prior_shapes_key="auto"` to use it (or, provide manually the column name and the `unassigned_value` argument).
        - If you have already run segmentation with Sopa, use `prior_shapes_key` to denote the name of the shapes (GeoDataFrame) containing the boundaries, e.g. `prior_shapes_key="cellpose_boundaries"`.

    Args:
        sdata: A `SpatialData` object.
        patch_width: Width of the patches, in microns. If `None`, creates only one patch.
        patch_overlap: Number of microns of overlap between patches.
        points_key: Optional key of the points on which the patches will be made. If not provided, it is found automatically.
        prior_shapes_key: Optional key of `sdata` containing the shapes with the prior segmentation, or column of the points dataframe. If `"auto"`, use the prior column from the technology.
        unassigned_value: If `prior_shapes_key` has been provided and corresponds to a points column: this argument is the value given to the transcript that are not inside any cell.
        min_points_per_patch: Minimum number of points/transcripts for a patch to be considered for segmentation.
        write_cells_centroids: If `True`, the centroids of the prior cells will be saved. This is useful for some segmentation tools such as ComSeg.
        roi_key: Optional name of the shapes that need to touch the patches. Patches that do not touch any shape will be ignored during segmentation. By default, uses `"region_of_interest"` if existing. If `None`, all patches will be used.
        key_added: Optional name of the patches to be saved. By default, uses `"transcripts_patches"`.
        **kwargs: Additional arguments passed to the `OnDiskTranscriptPatches` class.
    """
    assert not write_cells_centroids or prior_shapes_key, "write_cells_centroids argument requires prior_shapes_key"

    points_key, _ = get_spatial_element(
        sdata.points,
        key=points_key or sdata.attrs.get(SopaAttrs.TRANSCRIPTS),
        return_key=True,
    )

    if prior_shapes_key == "auto":
        assert SopaAttrs.PRIOR_TUPLE_KEY in sdata.attrs, (
            f"prior_shapes_key='auto' requires a prior segmentation to be present in the points dataframe ('{SopaAttrs.PRIOR_TUPLE_KEY}' must be in `sdata.attrs`)."
        )
        prior_shapes_key, unassigned_value = sdata.attrs[SopaAttrs.PRIOR_TUPLE_KEY]

    patches = OnDiskTranscriptPatches(
        sdata,
        points_key,
        patch_width=patch_width,
        patch_overlap=patch_overlap,
        prior_shapes_key=prior_shapes_key,
        unassigned_value=unassigned_value,
        min_points_per_patch=min_points_per_patch,
        write_cells_centroids=write_cells_centroids,
        roi_key=roi_key,
        **kwargs,
    )

    patches.write()
    patches.add_shapes(key_added=key_added)



import torch


class HistoSSLFeatures(torch.nn.Module):
    def __init__(self):
        super().__init__()
        # weights retrieved from https://github.com/ozanciga/self-supervised-histopathology
        url = "https://nextcloud.centralesupelec.fr/s/Kn4zEGMdWgnj2ac/download/tenpercent_resnet12.pth"
        self.model = torch.hub.load_state_dict_from_url(url)
        self.model.fc = torch.nn.Sequential()
        self.model.eval()

    def forward(self, x):
        return self.model(x)



import logging
from pathlib import Path
from typing import Literal

from spatialdata import SpatialData

from ...constants import SopaAttrs
from .utils import _default_image_kwargs

log = logging.getLogger(__name__)


def merscope(
    path: str | Path,
    backend: Literal["dask_image", "rioxarray"] | None = None,
    z_layers: int | list[int] | None = 3,
    region_name: str | None = None,
    slide_name: str | None = None,
    cells_boundaries: bool = False,
    cells_table: bool = False,
    image_models_kwargs: dict | None = None,
    imread_kwargs: dict | None = None,
    **kwargs: int,
) -> SpatialData:
    """Read MERSCOPE data as a `SpatialData` object. For more information, refer to [spatialdata-io](https://spatialdata.scverse.org/projects/io/en/latest/generated/spatialdata_io.merscope.html).

    This function reads the following files:
        - `detected_transcripts.csv`: transcripts locations and names
        - all the images under the `images` directory
        - `images/micron_to_mosaic_pixel_transform.csv`: affine transformation

    Args:
        path: Path to the MERSCOPE directory containing all the experiment files
        backend: Either `"dask_image"` or `"rioxarray"` (the latter uses less RAM, but requires `rioxarray` to be installed). By default, uses `"rioxarray"` if and only if the `rioxarray` library is installed.
        z_layers: Indices of the z-layers to consider. Either one `int` index, or a list of `int` indices. If `None`, then no image is loaded. By default, only the middle layer is considered (that is, layer 3).
        region_name: Name of the region of interest, e.g., `'region_0'`. If `None` then the name of the `path` directory is used.
        slide_name: Name of the slide/run. If `None` then the name of the parent directory of `path` is used (whose name starts with a date).
        cells_boundaries: Whether to read cell boundaries (polygons).
        cells_table: Whether to read the cells table.
        image_models_kwargs: Keyword arguments passed to `spatialdata.models.Image2DModel`.
        imread_kwargs: Keyword arguments passed to `dask_image.imread.imread`.
        kwargs: Additional keyword arguments passed to [`spatialdata_io.merscope`](https://spatialdata.scverse.org/projects/io/en/latest/generated/spatialdata_io.merscope.html).

    Returns:
        A `SpatialData` object representing the MERSCOPE experiment
    """
    from spatialdata_io.readers.merscope import merscope as merscope_spatialdata_io

    image_models_kwargs, imread_kwargs = _default_image_kwargs(image_models_kwargs, imread_kwargs)

    sdata: SpatialData = merscope_spatialdata_io(
        path,
        backend=backend,
        z_layers=z_layers,
        region_name=region_name,
        slide_name=slide_name,
        image_models_kwargs=image_models_kwargs,
        imread_kwargs=imread_kwargs,
        cells_boundaries=cells_boundaries,
        cells_table=cells_table,
        **kwargs,
    )

    ### Add Sopa attributes to detect the spatial elements
    if z_layers is not None:
        if not isinstance(z_layers, int) and len(z_layers) == 1:
            z_layers = z_layers[0]
        if isinstance(z_layers, int):
            for key in sdata.images:
                if key.endswith(f"_z{z_layers}"):
                    sdata.attrs[SopaAttrs.CELL_SEGMENTATION] = key
        else:
            log.warning(
                f"Multiple z-layers provided: {z_layers}. Not deciding which image should be used for cell segmentation."
            )

    for key in sdata.points:
        if key.endswith("_transcripts"):
            sdata.attrs[SopaAttrs.TRANSCRIPTS] = key

            if "cell_id" in sdata.points[key].columns:
                sdata.attrs[SopaAttrs.PRIOR_TUPLE_KEY] = ["cell_id", -1]
            break

    return sdata



from ._baysor import baysor
from ._cellpose import cellpose, cellpose_patch
from ._comseg import comseg
from ._custom import custom_staining_based
from ._dummy import dummy_method
from ._proseg import proseg
from ._stardist import stardist, stardist_patch



class SopaKeys:
    # Segmentation keys
    CELLPOSE_BOUNDARIES = "cellpose_boundaries"
    BAYSOR_BOUNDARIES = "baysor_boundaries"
    COMSEG_BOUNDARIES = "comseg_boundaries"
    CUSTOM_BOUNDARIES = "custom_boundaries"
    STARDIST_BOUNDARIES = "stardist_boundaries"
    PROSEG_BOUNDARIES = "proseg_boundaries"

    # Patches keys
    PATCHES = "image_patches"
    TRANSCRIPTS_PATCHES = "transcripts_patches"
    EMBEDDINGS_PATCHES = "embeddings_patches"
    BOUNDS = "bboxes"
    PATCHES_ILOCS = "ilocs"
    ROI = "region_of_interest"
    PRIOR_SHAPES_KEY = "prior_shapes_key"
    SOPA_PRIOR = "sopa_prior"
    POINTS_KEY = "points_key"

    # Other SpatialData keys
    TABLE = "table"
    OLD_TABLE_PREFFIX = "old_"

    # Table keys
    UNS_KEY = "sopa_attrs"
    UNS_HAS_TRANSCRIPTS = "transcripts"
    UNS_HAS_INTENSITIES = "intensities"
    UNS_CELL_TYPES = "cell_types"
    INTENSITIES_OBSM = "intensities"
    REGION_KEY = "region"
    SLIDE_KEY = "slide"
    INSTANCE_KEY = "cell_id"
    DEFAULT_CELL_KEY = "cell"
    ORIGINAL_AREA_OBS = "baysor_area"
    CELL_OVERLAY_KEY = "is_overlay"
    AREA_OBS = "area"
    Z_SCORES = "z_scores"

    # Geometry keys
    GEOMETRY_AREA = "area"
    GEOMETRY_LENGTH = "length"
    GEOMETRY_ROUNDNESS = "roundness"
    GEOMETRY_COUNT = "n_components"

    # Misc keys
    LOW_QUALITY_TRANSCRIPT_KEY = "low_quality_transcript"


class SopaAttrs:
    CELL_SEGMENTATION = "cell_segmentation_image"
    TISSUE_SEGMENTATION = "tissue_segmentation_image"
    BINS_TABLE = "bins_table"
    TRANSCRIPTS = "transcripts_dataframe"
    BOUNDARIES = "boundaries_shapes"
    UID = "sopa_uid"
    XENIUM_OUTPUT_PATH = "xenium_output_path"
    PRIOR_TUPLE_KEY = "transcript_to_cell_assignment"


class SopaFiles:
    SOPA_CACHE_DIR = ".sopa_cache"
    TRANSCRIPT_CACHE_DIR = "transcript_patches"
    PATCHES_FILE_IMAGE = "patches_file_image"
    PATCHES_FILE_TRANSCRIPTS = "patches_file_transcripts"
    TRANSCRIPTS_FILE = "transcripts.csv"
    CENTROIDS_FILE = "centroids.csv"
    JSON_CONFIG_FILE = "config.json"
    TOML_CONFIG_FILE = "config.toml"


VALID_DIMENSIONS = ("c", "y", "x")
LOW_AVERAGE_COUNT = 0.01
ATTRS_KEY = "spatialdata_attrs"



import logging
import warnings
from pathlib import Path

import dask.array as da
from dask_image.imread import imread
from spatialdata import SpatialData
from spatialdata.models import Image2DModel
from xarray import DataArray

from ...constants import SopaAttrs
from .utils import _clip_intensity_values, _default_image_kwargs

log = logging.getLogger(__name__)


def hyperion(path: Path, image_models_kwargs: dict | None = None, imread_kwargs: dict | None = None) -> SpatialData:
    """Read Hyperion data as a `SpatialData` object

    Args:
        path: Path to the directory containing the Hyperion `.tiff` images
        image_models_kwargs: Keyword arguments passed to `spatialdata.models.Image2DModel`.
        imread_kwargs: Keyword arguments passed to `dask_image.imread.imread`.

    Returns:
        A `SpatialData` object with a 2D-image of shape `(C, Y, X)`
    """
    image_models_kwargs, imread_kwargs = _default_image_kwargs(image_models_kwargs, imread_kwargs)

    files = [file for file in Path(path).iterdir() if file.suffix == ".tiff"]

    names = _get_channel_names_hyperion(files)
    image = da.concatenate(
        [imread(file, **imread_kwargs) for file in files],
        axis=0,
    )

    image = image.rechunk(chunks=image_models_kwargs["chunks"])

    log.info(f"Found channel names {names}")

    image_name = Path(path).absolute().stem

    image = DataArray(image, dims=["c", "y", "x"], name=image_name, coords={"c": names})
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        image = _clip_intensity_values(image)

    image = Image2DModel.parse(
        image,
        c_coords=image.coords["c"].values,
        **image_models_kwargs,
    )

    return SpatialData(images={image_name: image}, attrs={SopaAttrs.CELL_SEGMENTATION: image_name})


def _get_channel_names_hyperion(files: list[Path]):
    return [file.name[:-9].split("_")[1] for file in files]



import logging
from enum import Enum

import geopandas as gpd
import numpy as np
import skimage
from shapely.geometry import GeometryCollection, box
from skimage.morphology import disk
from spatialdata import SpatialData
from spatialdata.models import ShapesModel
from xarray import DataArray, DataTree

from ..constants import SopaAttrs, SopaKeys
from ..shapes import _vectorize_mask, expand_radius, to_valid_polygons
from ..utils import add_spatial_element, copy_transformations, get_spatial_element

log = logging.getLogger(__name__)


def tissue(
    sdata: SpatialData,
    image_key: str | None = None,
    level: int = -1,
    mode: str | None = None,
    expand_radius_ratio: float = 0.05,
    channel: str | None = None,
    clip_parameters: tuple[float, float] = (0.9, 5),
    blur_kernel_size: int = 5,
    open_kernel_size: int = 5,
    close_kernel_size: int = 5,
    drop_threshold: float = 0.01,
    allow_holes: bool = True,
    key_added: str = SopaKeys.ROI,
):
    """Perform a contouring of the tissue (i.e., "tissue-segmentation"). The resulting regions-of-interest(s) are saved as shapes in the `SpatialData` object. There are two
    modes available: `saturation` and `staining`. The `saturation` mode is used for H&E data, while the `staining` mode
    is used on staining images (more details below).

    !!! info "Saturation mode"
        This segmentation method first transforms the image from RBG color space to HSV. Then,
        on the basis of the saturation channel, a median blurring is applied with an element of size `blur_kernel_size`
        before running the Otsu method. Then a morphological opening and closing are applied as a prostprocessing
        step with square elements of size `open_kernel_size` and `close_kernel_size`. Lastly, the connected components
        with size less than `drop_threshold * number_of_pixel_of_the_image` are removed, and the
        rest are converted into polygons.

    !!! info "Staining mode"
        Instead of extracting the saturation channel, the image is converted to a grayscale image by taking the maximum
        value of all channels (or the specified channel, if `"channel"` is given). The rest of the steps are the same as in the saturation mode.

    Args:
        sdata: A `SpatialData` object representing an H&E image
        image_key: Optional key of the H&E image
        level: Level of the multiscale image on which the segmentation will be performed (if the image is a `DataTree`)
        mode: Two modes are available: `saturation` (for H&E data) and `staining`. By default, `saturation` is used only if there are exactly 3 channels.
        expand_radius_ratio: The ratio of the radius of the polygons that will be expanded.
        channel: The channel to use for the `staining` mode. If `None`, the maximum value of all channels is used.
        clip_parameters: Parameters used to get the threshold used to clip the image before converting it to an 8-bit image (only used in "staining" mode). The first parameter is the quantile, and the second is the divisor. By default, the threshold is the 90th quantile divided by 5.
        blur_kernel_size: The kernel size of the median bluring operation
        open_kernel_size: The kernel size of the morphological openning operation
        close_kernel_size: The kernel size of the morphological closing operation
        drop_threshold: Segments that cover less area than a ratio of `drop_threshold` of the number of pixels of the image will be removed
        allow_holes: If `True`, the holes in the polygons will be kept. If `False`, the holes will be removed.
        key_added: Name of the spatial element that will be added, containing the segmented tissue polygons.
    """
    image, mode = _get_image_and_mode(sdata, image_key, mode, channel)

    if key_added in sdata.shapes:
        log.warning(f"sdata['{key_added}'] was already existing, but tissue segmentation is run on top")

    if isinstance(image, DataTree):
        level_keys = list(image.keys())
        image: DataArray = next(iter(image[level_keys[level]].values()))

    geo_df = TissueSegmentation(
        image=image,
        blur_kernel_size=blur_kernel_size,
        open_kernel_size=open_kernel_size,
        close_kernel_size=close_kernel_size,
        drop_threshold=drop_threshold,
        channel=channel,
        clip_parameters=clip_parameters,
        allow_holes=allow_holes,
    ).get_polygons(mode)

    if not len(geo_df):
        log.warning(
            "No polygon has been found after tissue segmentation. "
            "Check that there is some tissue in the image, or consider updating the function parameters."
        )
        return

    geo_df = expand_radius(geo_df, expand_radius_ratio)
    geo_df = geo_df.explode(index_parts=False, ignore_index=True)
    geo_df = to_valid_polygons(geo_df, simple_polygon=not allow_holes)

    geo_df = ShapesModel.parse(geo_df, transformations=copy_transformations(image))

    add_spatial_element(sdata, key_added, geo_df)


class AvailableModes(Enum):
    SATURATION = "saturation"
    STAINING = "staining"

    @classmethod
    def check_available(cls, mode: str | None):
        if mode is not None:
            available_modes = [c.value for c in cls]
            assert mode in available_modes, f"Mode '{mode}' not available. Available modes are {available_modes}"


class TissueSegmentation:
    SIZE_THRESHOLD_LARGE_IMAGE = 1e8

    def __init__(
        self,
        image: DataArray,
        blur_kernel_size: int,
        open_kernel_size: int,
        close_kernel_size: int,
        drop_threshold: float,
        channel: str | None,
        clip_parameters: tuple[float, float],
        allow_holes: bool,
    ):
        self.image = image
        self.blur_kernel_size = blur_kernel_size
        self.open_kernel_size = open_kernel_size
        self.close_kernel_size = close_kernel_size
        self.drop_threshold = drop_threshold
        self.channel = channel
        self.clip_parameters = clip_parameters
        self.allow_holes = allow_holes

        if image.sizes["y"] * image.sizes["x"] > self.SIZE_THRESHOLD_LARGE_IMAGE:
            log.warning(
                "Tissue segmentation is computationally expensive for large images. "
                "Consider using a smaller image, or set the `level` parameter."
            )

    def get_polygons(self, mode: str) -> gpd.GeoDataFrame:
        thumbnail_2d = getattr(self, mode)()

        return self.otsu(thumbnail_2d)

    def staining(self) -> np.ndarray:
        if self.channel is None:
            thumbnail_2d = np.array(self.image.max(dim="c").transpose("y", "x"))
        else:
            thumbnail_2d = np.array(self.image.sel(c=self.channel).transpose("y", "x"))

        threshold = np.quantile(thumbnail_2d, self.clip_parameters[0]) / self.clip_parameters[1]
        thumbnail_2d = ((thumbnail_2d / threshold).clip(0, 1) * 255).astype(np.uint8)

        return thumbnail_2d

    def saturation(self) -> np.ndarray:
        assert self.image.sizes["c"] == 3, "The image should be in RGB color space"

        thumbnail = np.array(self.image.transpose("y", "x", "c"))

        assert thumbnail.dtype == np.uint8, "In 'saturation' mode, the image should have the uint8 dtype"

        thumbnail = (skimage.color.rgb2hsv(thumbnail) * 255).astype("uint8")
        return thumbnail[:, :, 1]  # saturation channel

    def otsu(self, thumbnail_2d: np.ndarray) -> gpd.GeoDataFrame:
        """Peform tissue segmentation using Otsu's method.

        Args:
            thumbnail_2d: A low-resolution 2D image (YX dimensions) of the tissue.

        Returns:
            A GeoDataFrame containing the segmented polygon(s).
        """
        thumbnail_blurred = skimage.filters.median(thumbnail_2d, footprint=disk(self.blur_kernel_size))

        mask = thumbnail_blurred > skimage.filters.threshold_otsu(thumbnail_blurred)

        mask_open = skimage.morphology.opening(mask, footprint=disk(self.open_kernel_size)).astype(np.uint8)
        mask_open_close = skimage.morphology.closing(mask_open, footprint=disk(self.close_kernel_size)).astype(np.uint8)

        labels = skimage.measure.label(mask_open_close, connectivity=2)

        geo_df = _vectorize_mask(labels, allow_holes=self.allow_holes)
        return gpd.GeoDataFrame(geometry=[geo_df.union_all()])


def _get_image_and_mode(
    sdata: SpatialData, image_key: str | None, mode: str | None, channel: str | None
) -> tuple[DataArray | DataTree, str]:
    AvailableModes.check_available(mode)
    assert channel is None or mode != AvailableModes.SATURATION.value, (
        "The `channel` argument is only used in the `staining` mode"
    )

    if channel is not None:
        mode = AvailableModes.STAINING.value

    if mode == AvailableModes.STAINING.value:
        image = get_spatial_element(sdata.images, key=image_key or sdata.attrs.get(SopaAttrs.CELL_SEGMENTATION))
        return image, mode

    if mode == AvailableModes.SATURATION.value:
        image = get_spatial_element(sdata.images, key=image_key or sdata.attrs.get(SopaAttrs.TISSUE_SEGMENTATION))
        return image, mode

    if image_key is not None:
        image = get_spatial_element(sdata.images, key=image_key)
        return image, _get_default_mode(image)

    # now, we have no image_key, no mode, no channel

    image_key, image = get_spatial_element(
        sdata.images,
        key=sdata.attrs.get(SopaAttrs.TISSUE_SEGMENTATION) or sdata.attrs.get(SopaAttrs.CELL_SEGMENTATION),
        return_key=True,
    )
    mode = _get_default_mode(image)

    log.info(f"Using {image_key=} and {mode=} as default")

    return image, mode


def _get_default_mode(image: DataArray | DataTree) -> str:
    n_channels = image.sizes["c"] if isinstance(image, DataArray) else next(iter(image.values())).sizes["c"]
    return AvailableModes.SATURATION.value if n_channels == 3 else AvailableModes.STAINING.value


def shapes_bounding_box(sdata: SpatialData, shape_key: str, key_added: str = SopaKeys.ROI):
    """Compute the bounding box of a shape layer and save it as a new shape in the `SpatialData` object.
    For instance, this can be used on VisiumHD data to use the bounding box of the bins as the ROI.

    Args:
        sdata: A `SpatialData` object
        shape_key: Key of the shape that will be used to compute the bounding box
        key_added: Name of the new shape that will be added to the `SpatialData` object
    """
    shapes = get_spatial_element(sdata.shapes, key=shape_key)

    if key_added in sdata.shapes:
        log.warning(f"sdata['{key_added}'] was already existing, but shapes_bounding_box is run on top")

    bounding_box = GeometryCollection(shapes.geometry.values)
    bounding_box = gpd.GeoDataFrame(geometry=[box(*bounding_box.bounds)])

    geo_df = ShapesModel.parse(bounding_box, transformations=copy_transformations(shapes))

    add_spatial_element(sdata, key_added, geo_df)



import logging

from xarray import DataArray, DataTree

log = logging.getLogger(__name__)


class ReaderBase:
    path: str
    name = "base"
    slide = None

    def read_region(self, location, level, size, **kwargs):
        raise NotImplementedError

    def get_metadata(self):
        """Get metadata from the slide."""
        return {
            "properties": self.properties,
            "dimensions": self.dimensions,
            "level_count": self.level_count,
            "level_dimensions": self.level_dimensions,
            "level_downsamples": self.level_downsamples,
        }

    def get_zarr_store(self):
        """Get the Zarr store for the slide."""
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    @property
    def properties(self):
        raise NotImplementedError

    @property
    def level_downsamples(self):
        raise NotImplementedError

    @property
    def level_count(self):
        raise NotImplementedError

    @property
    def level_dimensions(self):
        raise NotImplementedError

    @property
    def dimensions(self):
        raise NotImplementedError


class OpenSlideReader(ReaderBase):
    """Reader using the openslide backend."""

    name = "openslide"

    def __init__(self, path: str):
        try:
            from openslide import OpenSlide
        except ImportError:
            raise ImportError(
                "To use the openslide backend, you need to install it, e.g.,: `pip install openslide-python openslide-bin`."
            )

        self.path = path
        self.slide = OpenSlide(path)

    def read_region(self, location, level, size):
        """Get a region from the slide."""
        return self.slide.read_region(location, level, size).convert("RGB")

    def get_zarr_store(self, tilesize: int = 512):
        from zarr.storage import KVStore

        from ._wsi_store import WsiStore

        return KVStore(WsiStore(self, tilesize))

    def close(self):
        """Close the slide."""
        self.slide.close()

    @property
    def properties(self):
        """Get the properties of the slide."""
        return self.slide.properties

    @property
    def level_downsamples(self):
        return self.slide.level_downsamples

    @property
    def level_count(self):
        return len(self.slide.level_downsamples)

    @property
    def level_dimensions(self):
        return self.slide.level_dimensions

    @property
    def dimensions(self):
        """Get the dimensions of the slide."""
        return self.slide.dimensions


class TiffSlideReader(ReaderBase):
    """Reader using the tiffslide backend."""

    name = "tiffslide"

    def __init__(self, path: str):
        import tiffslide

        self.path = path
        self.slide = tiffslide.open_slide(path)

    def read_region(self, location, level, size):
        """Get a region from the slide."""
        return self.slide.read_region(location, level, size).convert("RGB")

    def get_zarr_store(self, tilesize: int = 512):
        return self.slide.zarr_group.store

    def close(self):
        """Close the slide."""
        self.slide.close()

    @property
    def properties(self):
        """Get the properties of the slide."""
        return self.slide.properties

    @property
    def level_downsamples(self):
        return self.slide.level_downsamples

    @property
    def level_count(self):
        return len(self.slide.level_downsamples)

    @property
    def level_dimensions(self):
        return self.slide.level_dimensions

    @property
    def dimensions(self):
        """Get the dimensions of the slide."""
        return self.slide.dimensions


class SlideIOReader(ReaderBase):
    """Reader using the SlideIO backend."""

    name = "slideio"

    def __init__(self, path: str):
        import dask

        dask.config.set(scheduler="single-threaded")
        log.warning("SlideIOReader is not multi-threaded compatible, setting dask scheduler to single-threaded.")

        try:
            import slideio
        except ImportError:
            raise ImportError("To use the slideio backend, you need to install it, e.g.,: `pip install slideio`.")

        self.path = path
        self.slide = slideio.open_slide(path)

    def read_region(self, location, level, size):
        import numpy as np
        from PIL import Image

        with self.slide.get_scene(0) as scene:
            scaling = 1 / scene.get_zoom_level_info(level).scale
            (x, y) = location
            x_end = min(x + int(scaling * size[0]), scene.size[0])
            y_end = min(y + int(scaling * size[1]), scene.size[1])
            x_width = x_end - x
            y_height = y_end - y
            tile_x = int(np.round(x_width / scaling))
            tile_y = int(np.round(y_height / scaling))
            _tile = scene.read_block((x, y, x_width, y_height), (tile_x, tile_y))
            tile = np.zeros((size[1], size[0], 3), dtype=np.uint8)
            tile[: _tile.shape[0], : _tile.shape[1], :] = _tile[:, :, :3]  # Ensure RGB format
        return Image.fromarray(tile)

    def get_zarr_store(self, tilesize: int = 512):
        from zarr.storage import KVStore

        from ._wsi_store import WsiStore

        return KVStore(WsiStore(self, tilesize))

    def close(self):
        pass

    @property
    def properties(self):
        return {"slideio.objective-power": self.slide.get_scene(0).magnification}

    @property
    def level_downsamples(self):
        return [
            1 / self.slide.get_scene(0).get_zoom_level_info(i).scale
            for i in range(self.slide.get_scene(0).num_zoom_levels)
        ]

    @property
    def level_count(self):
        return self.slide.get_scene(0).num_zoom_levels

    @property
    def level_dimensions(self):
        return [
            (
                self.slide.get_scene(0).get_zoom_level_info(i).size.width,
                self.slide.get_scene(0).get_zoom_level_info(i).size.height,
            )
            for i in range(self.slide.get_scene(0).num_zoom_levels)
        ]

    @property
    def dimensions(self):
        """Get the dimensions of the slide."""
        return self.slide.get_scene(0).size


class XarrayReader(ReaderBase):
    """Reader for xarray data (no backend used)."""

    name = "xarray"

    def __init__(self, slide: DataArray | DataTree):
        self.slide = slide

    def read_region(self, location, level, size):
        """Get a region from the slide."""
        x_start, y_start = location
        x_end, y_end = x_start + size[0], y_start + size[1]
        image = self.slide[f"scale{level}"].image if isinstance(self.slide, DataTree) else self.slide
        tile = image[:, slice(y_start, y_end), slice(x_start, x_end)]
        return tile.transpose("y", "x", "c")

    def close(self):
        """Close the slide."""
        self.slide.close()


def get_reader(backend: str) -> type[ReaderBase]:
    """Get a reader for the specified backend."""
    readers = {
        "openslide": OpenSlideReader,
        "tiffslide": TiffSlideReader,
        "slideio": SlideIOReader,
        "xarray": XarrayReader,
    }
    if backend not in readers:
        raise ValueError(f"Unknown backend: {backend}. Supported backends are: {', '.join(readers.keys())}.")
    return readers[backend]



import logging

import geopandas as gpd
import numpy as np
import shapely
from shapely.errors import GEOSException

from .validate import ensure_polygon

log = logging.getLogger(__name__)


def expand_radius(
    geo_df: gpd.GeoDataFrame, expand_radius_ratio: float | None, no_overlap: bool = False, inplace: bool = False
) -> gpd.GeoDataFrame:
    """Expand the radius of the cells by a given ratio.

    Args:
        geo_df: A GeoDataFrame containing the cells or shapes.
        expand_radius_ratio: Ratio to expand the cells polygons for channels averaging. For instance, a ratio of 0.5 expands the shape radius by 50%. If `None`, doesn't expand cells.
        no_overlap: *Experimental feature*: if `True`, ensures that the expanded cells do not overlap by computing the Voronoi diagram of the centroids of the cells.
        inplace: If `True`, modifies the input GeoDataFrame in place. If `False`, returns a new GeoDataFrame.

    Returns:
        A GeoDataFrame with the expanded cells.
    """
    if not inplace:
        geo_df = geo_df.copy()

    if not expand_radius_ratio:
        return geo_df

    expand_radius_ = expand_radius_ratio * np.mean(np.sqrt(geo_df.area / np.pi))
    geo_df.geometry = geo_df.buffer(expand_radius_)

    if no_overlap:
        log.warning(
            "Computing Voronoi polygons to ensure no overlap between shapes is still experimental. It can take 10+ minutes for 100k+ shapes."
        )
        geo_df.geometry = remove_overlap(geo_df, as_gdf=False)

    return geo_df


def remove_overlap(geo_df: gpd.GeoDataFrame, as_gdf: bool = True) -> gpd.GeoSeries | gpd.GeoDataFrame:
    """Remove overlapping areas from a GeoDataFrame by computing the Voronoi polygons of the shapes.

    Args:
        geo_df: A GeoDataFrame containing the shapes.
        as_gdf: Whether to return a GeoDataFrame or a GeoSeries.

    Returns:
        A GeoSeries or GeoDataFrame with the overlapping areas removed.
    """
    geo_df["_index"] = geo_df.index  # to keep track of the index after the overlay

    overlay = geo_df.overlay(geo_df, how="intersection")
    overlap = overlay[overlay["_index_1"] != overlay["_index_2"]].union_all()

    del geo_df["_index"]

    if overlap.is_empty:
        return geo_df.geometry

    shapes_no_overlap = geo_df.difference(overlap).buffer(-1e-4)  # to avoid touching polygons on single points
    _voronoi = voronoi_frames(shapes_no_overlap)

    geometry = geo_df.intersection(_voronoi)

    nan_locs = geometry.type.isna()

    if nan_locs.any():
        log.warning(f"Found {nan_locs.sum()} NaN geometries after removing the overlap. Replacing with empty polygons.")
        geometry[nan_locs] = shapely.Polygon()

    geometry = geometry.buffer(1e-4)  # to re-expand the polygons to their original size
    geometry = geometry.map(ensure_polygon)

    if not as_gdf:
        return geometry

    geo_df = geo_df.copy()
    geo_df.geometry = geometry
    return geo_df


def voronoi_frames(
    geometry: gpd.GeoSeries | gpd.GeoDataFrame,
    clip: str | shapely.Geometry | None = "bounding_box",
    grid_size: float = 1e-5,
) -> gpd.GeoSeries:
    """
    Copied and simplified from https://pysal.org/libpysal/generated/libpysal.cg.voronoi_frames.html
    """
    # Check if the input geometry is in a geographic CRS
    if geometry.crs and geometry.crs.is_geographic:
        raise ValueError(
            "Geometry is in a geographic CRS. "
            "Use 'GeoSeries.to_crs()' to re-project geometries to a "
            "projected CRS before using voronoi_polygons.",
        )

    # Set precision of the input geometry (avoids GEOS precision issues)
    objects: gpd.GeoSeries = shapely.set_precision(geometry.geometry.copy(), grid_size)

    geom_types = objects.geom_type
    assert geom_types.isin(["Polygon", "MultiPolygon"]).all(), (
        "Only Polygon and MultiPolygon geometries are supported to remove overlaps."
    )

    limit = _get_limit(objects, clip)

    # Compute Voronoi polygons
    voronoi = shapely.voronoi_polygons(shapely.GeometryCollection(objects.values), extend_to=limit)
    # Get individual polygons out of the collection
    polygons = gpd.GeoSeries(shapely.make_valid(shapely.get_parts(voronoi)), crs=geometry.crs)

    # temporary fix for libgeos/geos#1062
    if not (polygons.geom_type == "Polygon").all():
        polygons = polygons.explode(ignore_index=True)
        polygons = polygons[polygons.geom_type == "Polygon"]

    ids_objects, ids_polygons = polygons.sindex.query(objects, predicate="intersects")

    # Dissolve polygons
    polygons = polygons.iloc[ids_polygons].groupby(objects.index.take(ids_objects)).agg(_union_with_fallback)
    if geometry.crs is not None:
        polygons = polygons.set_crs(geometry.crs)

    # ensure validity as union can occasionally produce invalid polygons that may
    # break the intersection below
    if not polygons.is_valid.all():
        polygons = polygons.make_valid()

    # Clip polygons if limit is provided
    if limit is not None:
        to_be_clipped = polygons.sindex.query(limit.boundary, "intersects")
        polygons.iloc[to_be_clipped] = polygons.iloc[to_be_clipped].intersection(limit)

    return polygons


def _union_with_fallback(arr):
    """
    Coverage union is finnicky with floating point precision and tends to occasionally
    raise an error from within GEOS we have no control over. It is not a data issue
    typically. Falling back to unary union if that happens.
    """
    try:
        r = shapely.coverage_union_all(arr)
    except GEOSException:
        r = shapely.union_all(arr)

    return r


def _get_limit(points, clip: shapely.Geometry | str | None | bool) -> shapely.Geometry | None:
    if isinstance(clip, shapely.Geometry):
        return clip
    if clip is None or clip is False:
        return None
    if clip.lower() == "bounding_box":
        return shapely.box(*points.total_bounds)
    raise ValueError(
        f"Clip type '{clip}' not understood. Try one of the supported options: [None, 'bounding_box', shapely.Polygon]."
    )



import logging
import re
from pathlib import Path

import dask.array as da
import tifffile as tf
from dask.delayed import delayed
from dask_image.imread import imread
from spatialdata import SpatialData
from spatialdata.models import Image2DModel

from ...constants import SopaAttrs
from .utils import _deduplicate_names, _default_image_kwargs

log = logging.getLogger(__name__)


def phenocycler(
    path: str | Path, channels_renaming: dict | None = None, image_models_kwargs: dict | None = None
) -> SpatialData:
    """Read Phenocycler data as a `SpatialData` object

    Args:
        path: Path to a `.qptiff` file, or a `.tif` file (if exported from QuPath)
        channels_renaming: A dictionnary whose keys correspond to channels and values to their corresponding new name. Not all channels need to be renamed.
        image_models_kwargs: Keyword arguments passed to `spatialdata.models.Image2DModel`.

    Returns:
        A `SpatialData` object with a 2D-image of shape `(C, Y, X)`
    """
    image_models_kwargs, _ = _default_image_kwargs(image_models_kwargs)

    path = Path(path)
    image_name = path.absolute().stem

    if path.suffix == ".qptiff":
        with tf.TiffFile(path) as tif:
            series = tif.series[0]
            names = _deduplicate_names([_get_channel_name_qptiff(page.description) for page in series])

            delayed_image = delayed(lambda series: series.asarray())(tif)
            image = da.from_delayed(delayed_image, dtype=series.dtype, shape=series.shape)
    elif path.suffix == ".tif":
        image = imread(path)
        names = _get_IJ_channel_names(path)
    else:
        raise ValueError(f"Unsupported file extension {path.suffix}. Must be '.qptiff' or '.tif'.")

    names = _rename_channels(names, channels_renaming)
    image = image.rechunk(chunks=image_models_kwargs["chunks"])

    image = Image2DModel.parse(
        image,
        dims=("c", "y", "x"),
        c_coords=names,
        **image_models_kwargs,
    )

    return SpatialData(images={image_name: image}, attrs={SopaAttrs.CELL_SEGMENTATION: image_name})


def _get_channel_name_qptiff(description):
    import xml.etree.ElementTree as ET

    root = ET.fromstring(description)

    for xml_path in [".//Biomarker", ".//ExcitationFilter//Bands//Name"]:
        field = root.find(xml_path)
        if field is not None:
            return field.text

    return re.search(r"<Name>(.*?)</Name>", description).group(1)


def _get_IJ_channel_names(path: str) -> list[str]:
    with tf.TiffFile(path) as tif:
        default_names = [str(i) for i in range(len(tif.pages))]

        if len(tif.pages) > 1:
            ij_metadata_tag = tif.pages[0].tags.get("IJMetadata", None)

            if ij_metadata_tag and "Labels" in ij_metadata_tag.value:
                return ij_metadata_tag.value["Labels"]

            log.warning("Could not find channel names in IJMetadata.")
            return default_names

        log.warning("The TIF file does not have multiple channels.")
        return default_names


def _rename_channels(names: list[str], channels_renaming: dict | None = None):
    log.info(f"Found channel names {names}")
    if channels_renaming is not None and len(channels_renaming):
        log.info(f"Channels will be renamed by the dictionnary: {channels_renaming}")
        names = [channels_renaming.get(name, name) for name in names]
        log.info(f"New names are: {names}")
    return names



import torch

from .conch import ConchFeatures
from .dinov2 import DINOv2Features
from .dummy import DummyFeatures
from .histo_ssl import HistoSSLFeatures
from .hoptimus0 import HOPTIMUSFeatures
from .resnet import Resnet50Features

__all__ = [
    "ConchFeatures",
    "DINOv2Features",
    "DummyFeatures",
    "HOPTIMUSFeatures",
    "HistoSSLFeatures",
    "Resnet50Features",
]

available_models: dict[str, type[torch.nn.Module]] = {
    "conch": ConchFeatures,
    "dinov2": DINOv2Features,
    "dummy": DummyFeatures,
    "hoptimus0": HOPTIMUSFeatures,
    "histo_ssl": HistoSSLFeatures,
    "resnet50": Resnet50Features,
}



import logging
import os
from typing import Callable

import dask
import dask.delayed
from dask.distributed import Client, progress
from tqdm import tqdm

log = logging.getLogger(__name__)


class Settings:
    ### General
    auto_save_on_disk: bool = True

    ### Parallelization
    _parallelization_backend: str | None = None
    available_parallelization_backends = [None, "dask"]
    dask_client_kwargs: dict = {}

    ### Backend settings
    native_read_region: bool = True

    ### Segmentation or aggregation
    gene_exclude_pattern: str | None = (
        "nan|<NA>|.*control.*|blank.*|antisense.*|unassigned.*|deprecated.*|intergenic.*|false.*|neg.*"
    )
    simplification_tolerance: float | None = None

    def __init__(self):
        self.parallelization_backend = os.environ.get("SOPA_PARALLELIZATION_BACKEND", None)

    @property
    def parallelization_backend(self):
        return self._parallelization_backend

    @parallelization_backend.setter
    def parallelization_backend(self, value):
        assert value in self.available_parallelization_backends, (
            f"Invalid parallelization backend. Available options are: {self.available_parallelization_backends}"
        )
        self._parallelization_backend = value

    def _run_with_backend(self, functions: list[Callable]):
        assert len(functions) > 0, "No function to run"

        if len(functions) == 1 or self.parallelization_backend is None:
            if len(functions) > 1:
                log.warning(
                    "Running without parallelization backend can be slow. "
                    "Consider using a backend, e.g. via `sopa.settings.parallelization_backend = 'dask'`, "
                    "or `export SOPA_PARALLELIZATION_BACKEND=dask`."
                )
            return [f() for f in tqdm(functions)]

        log.info(f"Using {self.parallelization_backend} backend")
        return getattr(self, f"_run_{self.parallelization_backend}_backend")(functions)

    ### Dask backend
    def _run_dask_backend(self, functions: list[Callable]):
        import logging
        import os

        from distributed.scheduler import logger as _logger

        _logger.setLevel(logging.ERROR)

        n_cpus: int = os.cpu_count()
        dask_client_kwargs = self.dask_client_kwargs.copy()
        if "n_workers" not in dask_client_kwargs and len(functions) < n_cpus:
            dask_client_kwargs["n_workers"] = len(functions)

        with Client(**dask_client_kwargs) as client:
            ram_per_worker = client.cluster.workers[0].memory_manager.memory_limit

            if ram_per_worker < 4 * 1024**3:
                log.warning(
                    f"Each worker has less than 4GB of RAM ({ram_per_worker / 1024**3:.2f}GB), which may not be enough. "
                    f"Consider setting `sopa.settings.dask_client_kwargs['n_workers']` to use less workers ({len(client.cluster.workers)} currently)."
                )

            functions = [dask.delayed(function)() for function in functions]
            futures = dask.persist(*functions)
            progress(futures, notebook=False)
            return client.gather(client.compute(futures))


settings = Settings()



from ._rasterize import rasterize
from .validate import ensure_polygon, to_valid_polygons
from .expand import expand_radius, remove_overlap
from ._vectorize import _vectorize_mask, pixel_outer_bounds, vectorize



import logging

import numpy as np
import pandas as pd
from anndata import AnnData
from spatialdata import SpatialData
from tqdm import tqdm

from ..constants import SopaKeys
from .build import _check_has_delaunay

log = logging.getLogger(__name__)


def cells_to_groups(
    adata: AnnData,
    group_key: str,
    key_added_prefix: str | None = None,
    ignore_zeros: bool = False,
) -> pd.DataFrame | None:
    """Compute the hop-distance between each cell and a cell category/group.

    Args:
        adata: An `AnnData` object, or a `SpatialData object`
        group_key: Key of `adata.obs` containing the groups
        key_added_prefix: Prefix to the key added in `adata.obsm`. If `None`, will return the `DataFrame` instead of saving it.
        ignore_zeros: If `True`, a cell distance to its own group is not 0, but the distance to the closest other cell of the same group.

    Returns:
        A `Dataframe` of shape `n_obs * n_groups`, or `None` if `key_added_prefix` was provided (in this case, the dataframe is saved in `"{key_added_prefix}{group_key}"`)
    """
    if isinstance(adata, SpatialData):
        adata = adata.tables[SopaKeys.TABLE]

    _check_has_delaunay(adata)

    distances_to_groups = {}

    if adata.obs[group_key].dtype.name != "category":
        log.info(f"Converting adata.obs['{group_key}'] to category")
        adata.obs[group_key] = adata.obs[group_key].astype("category")

    for group_id in tqdm(adata.obs[group_key].cat.categories):
        group_nodes = np.where(adata.obs[group_key] == group_id)[0]

        distances = np.full(adata.n_obs, np.nan)

        if not ignore_zeros:
            distances[group_nodes] = 0
            visited = set(group_nodes)
        else:
            visited = set()

        queue = group_nodes
        current_distance = 0

        while len(queue):
            distances[queue] = current_distance

            neighbors = set(adata.obsp["spatial_connectivities"][queue].indices)
            queue = np.array(list(neighbors - visited))
            visited |= neighbors

            current_distance += 1

        distances_to_groups[group_id] = distances

    df_distances = pd.DataFrame(distances_to_groups, index=adata.obs_names)

    if key_added_prefix is None:
        return df_distances
    adata.obsm[f"{key_added_prefix}{group_key}"] = df_distances


def mean_distance(
    adata: AnnData,
    group_key: str,
    target_group_key: str | None = None,
    ignore_zeros: bool = False,
    correction: bool = False,
    relative_change: bool = False,
) -> pd.DataFrame:
    """Mean distance between two groups (typically, between cell-types, or between cell-types and domains).

    !!! warning
        When computing distances for multiple slides, differences in cell-type proportions can introduce bias. For example, if group G is more abundant in one slide, the average distance from other groups to G will naturally appear smaller in that slide, simply due to higher local density. If you want to perform fair comparisons across samples, consider using the `correction` argument to adjust for this bias.

    Note:
        The distance is a number of hops, i.e. a distance of 10 between a pDC and a T cell means that there are 10 cells on the closest path from one to the other cell.

    Args:
        adata: An `AnnData` object, or a `SpatialData object`
        group_key: Key of `adata.obs` containing the groups
        target_group_key: Key of `adata.obs` containing the target groups (by default, uses `group_key`)
        ignore_zeros: If `True`, a cell distance to its own group is not 0, but the distance to the closest other cell of the same group.
        correction: If `True`, the computed distance is corrected by subtracting the expected distance under a null model where target labels are randomly distributed on a grid. This correction accounts for uneven label representation, which could otherwise deflate distances to more abundant target cell types.
        relative_change: Only used if `correction` is `True`. If `True`, the correction is applied as a relative change, i.e. the distance difference is also divided by the expected distance under the null model defined previously.

    Returns:
        `DataFrame` of shape `n_groups * n_groups_target` of mean hop-distances
    """
    if isinstance(adata, SpatialData):
        adata = adata.tables[SopaKeys.TABLE]

    target_group_key = group_key if target_group_key is None else target_group_key

    df_distances = cells_to_groups(adata, target_group_key, None, ignore_zeros=ignore_zeros)

    if ignore_zeros:
        df_distances.replace(0, np.nan, inplace=True)

    df_distances[group_key] = adata.obs[group_key]
    df_distances = df_distances.groupby(group_key, observed=False).mean()
    df_distances.columns.name = target_group_key

    if correction:
        log.info("Correcting the distances is still experimental, don't hesitate to report issues or feedback")
        proportions = adata.obs[target_group_key].value_counts(normalize=True)
        likelihood = proportions.map(lambda p: _random_distance_likelihood(p, ignore_zeros))

        df_distances -= likelihood
        if relative_change:
            df_distances /= likelihood

        if (not ignore_zeros) and (group_key == target_group_key):
            # distance to cell to same group is 0, we don't correct it
            df_distances.values[np.diag_indices_from(df_distances)] = 0

    return df_distances


def _random_distance_likelihood(proportion: float, ignore_zeros: bool, n_max: int = 30) -> float:
    """Compute the expected minimum between between a random cell and a cell-label with a given proportion of cells.

    It assumes cells are arranged in a grid, and the distance is the number of hops between two cells.

    E(D) = sum_{d=1}^{inf} d * ((1 - p)^{4*d(d-1)} - (1 - p)^{4*d(d+1)})
    """
    distances = np.arange(1, n_max)

    left = 4 * distances * (distances - 1) + int(not ignore_zeros)
    right = 4 * distances * (distances + 1) + int(not ignore_zeros)

    return (distances * ((1 - proportion) ** left - (1 - proportion) ** right)).sum()


def prepare_network(
    adata: AnnData,
    cell_type_key: str,
    niche_key: str,
    clip_weight: float = 3,
    node_colors: tuple[str] = ("#5c7dc4", "#f05541"),
    node_sizes: tuple[float | int] = (1.3, 5),
) -> tuple[pd.DataFrame, dict, dict, dict]:
    """Create a dataframe representing weights between cell-types and/or niches.
    This can be later use to plot a cell-type/niche represention of a whole slide
    using the netgraph library.

    Args:
        adata: An `AnnData` object
        cell_type_key: Key of `adata.obs` containing the cell types
        niche_key: Key of `adata.obs` containing the niches
        clip_weight: Maximum weight
        node_colors: Tuple of (cell-type color, niche color)
        node_sizes: Tuple of (cell-type size, niche size)

    Returns:
        A DataFrame of weights between cell-types and/or niches, and three dict for netgraph display
    """
    node_color, node_size, node_shape = {}, {}, {}

    log.info("Computing all distances for the 4 pairs of categories")
    weights = mean_distance(adata, cell_type_key)
    top_right = mean_distance(adata, cell_type_key, niche_key)
    bottom_left = mean_distance(adata, niche_key, cell_type_key)
    bottom_right = mean_distance(adata, niche_key, niche_key)

    for pop in weights.index:
        node_color[pop] = node_colors[0]
        node_size[pop] = node_sizes[0]
        node_shape[pop] = "o"

    for niche in bottom_right.index:
        node_color[niche] = node_colors[1]
        node_size[niche] = node_sizes[1]
        node_shape[niche] = "h"

    # assemble dataframe per-block
    bottom_left[bottom_right.columns] = bottom_right
    weights[top_right.columns] = top_right
    weights = pd.concat([weights, bottom_left], axis=0).copy()

    # convert distances to symmetric weights
    weights = 1 / weights
    np.fill_diagonal(weights.values, 0)
    weights = weights.clip(0, clip_weight)
    weights = (weights.T + weights) / 2

    return weights, node_color, node_size, node_shape



import logging

import dask.array as da
import dask.dataframe as dd
import geopandas as gpd
import numpy as np
import pandas as pd
from scipy.ndimage import gaussian_filter
from shapely.geometry import MultiPolygon, Point, Polygon, box
from spatialdata import SpatialData
from spatialdata.datasets import BlobsDataset
from spatialdata.models import Image2DModel, PointsModel, ShapesModel
from spatialdata.transformations import Affine, Identity, Scale

from ..constants import SopaAttrs, SopaKeys

log = logging.getLogger(__name__)


def toy_dataset(
    *_,
    length: int = 2_048,
    cell_density: float = 1e-4,
    n_points_per_cell: int = 100,
    c_coords: list[str] = ["DAPI", "CK", "CD3", "CD20"],  # noqa: B006
    genes: int | list[str] = ["EPCAM", "CD3E", "CD20", "CXCL4", "CXCL10"],  # noqa: B006
    sigma_factor: float = 0.05,
    pixel_size: float = 0.1,
    seed: int = 0,
    include_vertices: bool = False,
    include_image: bool = True,
    include_he_image: bool = True,
    apply_blur: bool = True,
    as_output: bool = False,
    transcript_cell_id_as_merscope: bool = False,
    add_nan_gene_name: bool = False,
    continuous_z_stack: bool = False,
    add_second_points_key: bool = False,
    points_3d: bool = False,
) -> SpatialData:
    """Generate a dummy dataset composed of cells generated uniformly in a square. It also has transcripts.

    Args:
        length: Size of the square, in pixels
        cell_density: Density of cells per pixel^2
        n_points_per_cell: Mean number of transcripts per cell
        c_coords: Channel names
        genes: Number of different genes, or list of gene names
        sigma_factor: Factor used to determine `sigma` for the gaussian blur.
        pixel_size: Number of microns in one pixel.
        seed: Numpy random seed
        include_vertices: Whether to include the vertices of the cells (as points) in the spatialdata object
        include_image: Whether to include the image in the spatialdata object
        apply_blur: Whether to apply gaussian blur on the image (without blur, cells are just one pixel)
        as_output: If `True`, the data will have the same format than an output of Sopa
        transcript_cell_id_as_merscope: If `True`, the cell IDs in the transcripts will start from 0, as in MERSCOPE
        add_nan_gene_name: If `True`, a NaN value will be added to the gene names for testing purposes
        continuous_z_stack: If `True`, the z-stack values will be continuous (not integers)
        add_second_points_key: If `True`, a second points key will be added to the dataset with dummy data for testing purposes
        points_3d: If `True`, the points will be 3D points instead of 2D points.

    Returns:
        A SpatialData object with a 2D image (`sdata["image"]`), the cells polygon boundaries (`sdata["cells"]`), the transcripts (`sdata["transcripts"]`), and optional cell vertices (`sdata["vertices"]`) if `include_vertices` is `True`.
    """
    np.random.seed(seed)

    grid_width = max(1, int(length * np.sqrt(cell_density)))
    dx = length / grid_width
    sigma = dx * sigma_factor
    n_cells = grid_width**2
    radius = int(dx) // 4
    cell_types_index = np.random.randint(0, max(1, len(c_coords) - 1), n_cells)

    log.info(
        f"Image of size {(len(c_coords), length, length)} with {n_cells} cells and {n_points_per_cell} transcripts per cell"
    )

    ### Compute cell vertices (xy array)
    vertices_x = dx / 2 + np.arange(grid_width) * dx
    x, y = np.meshgrid(vertices_x, vertices_x)
    xy = np.stack([x.ravel(), y.ravel()], axis=1)
    xy += np.random.uniform(-dx / 2, dx / 2, size=xy.shape)
    xy = xy.clip(0, length - 1).astype(int)

    vertices = pd.DataFrame(xy, columns=["x", "y"])

    ### Create images
    images = {}

    if include_image:
        x_circle, y_circle = _circle_coords(radius)

        image = np.zeros((len(c_coords), length, length))
        for i, (x, y) in enumerate(xy):
            y_coords = (y + y_circle).clip(0, image.shape[1] - 1)
            x_coords = (x + x_circle).clip(0, image.shape[2] - 1)
            image[0, y_coords, x_coords] = 1
            if len(c_coords) > 1:
                image[cell_types_index[i] + 1, y_coords, x_coords] = 1
        if apply_blur:
            image = gaussian_filter(image, sigma=sigma, axes=(1, 2))
        image = (image / image.max() * 255).astype(np.uint8)
        image = da.from_array(image, chunks=(1, 1024, 1024))
        images["image"] = Image2DModel.parse(image, c_coords=c_coords, dims=["c", "y", "x"])

    if include_he_image:
        he_image = _he_image(length // 2)
        scale = length / (length // 2)
        images["he_image"] = Image2DModel.parse(
            he_image,
            dims=["c", "y", "x"],
            transformations={"global": Scale([scale, scale], axes=["x", "y"])},
            scale_factors=[2, 2],
        )

    ### Create cell boundaries
    cells = [Point(vertex).buffer(radius).simplify(tolerance=1) for vertex in xy]
    bbox = box(0, 0, length - 1, length - 1)
    cells = [cell.intersection(bbox) for cell in cells]
    gdf = gpd.GeoDataFrame(geometry=cells)
    shapes_key = "cellpose_boundaries" if as_output else "cells"
    shapes = {shapes_key: ShapesModel.parse(gdf)}

    ### Create transcripts
    n_genes = n_cells * n_points_per_cell
    point_cell_index = np.arange(n_cells).repeat(n_points_per_cell)
    points_coords = radius / 2 * np.random.randn(n_genes, 2) + xy[point_cell_index]
    points_coords = points_coords.clip(0, length - 1)

    if isinstance(genes, int):
        gene_names = np.random.choice([f"gene_{i}" for i in range(genes)], size=n_genes)
    elif len(genes) and len(genes) >= len(c_coords) - 1:
        gene_names = np.full(n_genes, "", dtype="<U5")
        for i in range(len(genes)):
            where_cell_type = np.where(cell_types_index[point_cell_index] == i)[0]
            probabilities = np.full(len(genes), 0.2 / (len(genes) - 1))
            probabilities[i] = 0.8
            gene_names[where_cell_type] = np.random.choice(genes, len(where_cell_type), p=probabilities)
    else:
        gene_names = np.random.choice(genes, size=n_genes)

    gene_names = gene_names.astype(object)
    if add_nan_gene_name:
        gene_names[3] = np.nan  # Add a nan value for tests
        gene_names[4] = "blank"  # Add a blank value for tests

    z_stack = seed + np.random.randint(-1, 2, len(points_coords))
    if continuous_z_stack:
        z_stack = z_stack + np.random.randn(len(points_coords)) / 10

    df = pd.DataFrame({
        "x": points_coords[:, 0],
        "y": points_coords[:, 1],
        ("z" if points_3d else "z_stack"): z_stack,
        "genes": gene_names,
    })

    # apply an arbritrary transformation for a more complete test case
    affine = np.array([[pixel_size, 0, 100], [0, pixel_size, 600], [0, 0, 1]])
    df[["x", "y"]] = df[["x", "y"]] @ affine[:2, :2].T + affine[:2, 2]
    affine = Affine(affine, input_axes=["x", "y"], output_axes=["x", "y"]).inverse()

    df = dd.from_pandas(df, chunksize=2_000_000)

    points = {
        "transcripts": PointsModel.parse(
            df, transformations={"global": affine, "microns": Identity()}, feature_key="genes"
        ),
    }

    if add_second_points_key:
        misc_df = pd.DataFrame({"x": [0, 1], "y": [0, 1]})  # dummy dataframe for testing purposes
        points["misc"] = PointsModel.parse(misc_df, transformations={"global": Identity()})

    if include_vertices:
        points["vertices"] = PointsModel.parse(vertices)

    sdata = SpatialData(
        images=images,
        points=points,
        shapes=shapes,
        attrs={SopaAttrs.TRANSCRIPTS: "transcripts", SopaAttrs.PRIOR_TUPLE_KEY: ("cell_id", 0)},
    )

    if include_image:
        sdata.attrs[SopaAttrs.CELL_SEGMENTATION] = "image"

    if include_he_image:
        sdata.attrs[SopaAttrs.TISSUE_SEGMENTATION] = "he_image"

    from ..spatial import assign_transcript_to_cell

    assign_transcript_to_cell(sdata, "transcripts", shapes_key, "cell_id", unassigned_value=0)

    sdata["transcripts"]["cell_id"] = sdata["transcripts"]["cell_id"].astype(int) - int(transcript_cell_id_as_merscope)

    if as_output:
        _add_table(sdata)

    return sdata


def _add_table(sdata: SpatialData):
    from ..aggregation import Aggregator

    aggregator = Aggregator(sdata, shapes_key=SopaKeys.CELLPOSE_BOUNDARIES)

    aggregator.compute_table(gene_column="genes")


def _circle_coords(radius: int) -> tuple[np.ndarray, np.ndarray]:
    """Compute the coordinates of a circle

    Args:
        radius: Radius of the circle

    Returns:
        The x, y coordinates (two 1D ndarrays)
    """
    diameter = 2 * radius + 1
    mask = np.zeros((diameter, diameter), dtype=np.uint8)

    y, x = np.ogrid[: mask.shape[0], : mask.shape[1]]
    mask[((x - radius) ** 2 + (y - radius) ** 2) <= radius**2] = 1

    x_circle, y_circle = np.where(mask)
    return x_circle - radius, y_circle - radius


def _he_image(length: int) -> np.ndarray:
    from ..shapes import rasterize

    coords = np.array([
        [0.15, 0.15],
        [0.15, 0.7],
        [0.3, 0.8],
        [0.8, 0.85],
        [0.8, 0.8],
        [0.85, 0.12],
        [0.8, 0.2],
        [0.7, 0.4],
        [0.5, 0.5],
        [0.4, 0.5],
        [0.2, 0.35],
    ])

    circle_in = Point([0.5 * length, 0.2 * length]).buffer(0.1 * length)
    circle_out = Point([0.7 * length, 0.7 * length]).buffer(0.1 * length)
    polygon_he = Polygon(coords * length).buffer(0.1 * length)
    multi_polygon = MultiPolygon([polygon_he, circle_in])

    tissue = (rasterize(multi_polygon, [length, length]) - rasterize(circle_out, [length, length])).astype(np.uint8)

    image = np.random.randint(0, 50, size=(3, length, length), dtype=np.uint8)  # noise 1
    image[1] += (np.arange(length) / length * 10).astype(np.uint8)  # noise 2
    image[1] += (np.arange(length)[:, None] / length * 10).astype(np.uint8)  # noise 3
    image[1] += tissue * 60
    image[0] += tissue * 20
    image[2] += tissue * 40

    return 255 - image


def blobs(
    *_,
    length: int = 1_024,
    n_points: int = 10_000,
    c_coords=["DAPI", "CK", "CD3", "CD20"],  # noqa: B006
    **kwargs,
) -> SpatialData:
    """Adapts the blobs dataset from SpatialData for sopa. Please refer to the SpatialData documentation"""
    _blobs = BlobsDataset(length=length, n_points=n_points, c_coords=c_coords, n_channels=len(c_coords), **kwargs)

    image = _blobs._image_blobs(
        _blobs.transformations,
        _blobs.length,
        _blobs.n_channels,
        _blobs.c_coords,
    )
    image.data = (image.data * 255).astype(np.uint8)

    points = _blobs._points_blobs(_blobs.transformations, _blobs.length, _blobs.n_points)
    genes = pd.Series(np.random.choice(list("abcdef"), size=len(points))).astype("category")
    points["genes"] = dd.from_pandas(genes, npartitions=points.npartitions)

    return SpatialData(
        images={"blob_image": image},
        points={"blob_transcripts": points},
        attrs={SopaAttrs.CELL_SEGMENTATION: "blob_image", SopaAttrs.TRANSCRIPTS: "blob_transcripts"},
    )



import json
import logging
import os
import sys
import warnings
from functools import partial
from pathlib import Path

import geopandas as gpd
import numpy as np
from spatialdata import SpatialData

from ... import settings
from ...constants import SopaKeys
from ...utils import (
    delete_transcripts_patches_dirs,
    get_feature_key,
    get_transcripts_patches_dirs,
    set_boundaries_attrs,
    to_intrinsic,
)
from .utils import _check_transcript_patches, resolve

log = logging.getLogger(__name__)


def comseg(
    sdata: SpatialData,
    config: dict | str | None = None,
    min_area: float = 0,
    delete_cache: bool = True,
    recover: bool = False,
    key_added: str = SopaKeys.COMSEG_BOUNDARIES,
    patch_index: int | None = None,
):
    """Run [ComSeg](https://comseg.readthedocs.io/en/latest/) segmentation on a SpatialData object, and add a GeoDataFrame containing the cell boundaries.

    !!! warning "ComSeg installation"
        Make sure to install ComSeg (`pip install comseg`) for this method to work.

    !!! info "Transcript patches"
        To use ComSeg, make sure to run [sopa.make_transcript_patches][] with a `prior_shapes_key` and `write_cells_centroids=True`.

    Args:
        sdata: A `SpatialData` object.
        config: Optional configuration dictionary or path to a JSON file containing a valid ComSeg config. By default, a configuration is inferred based on the cell area of the prior segmentation.
        min_area: Minimal area (in microns^2) of a cell to be considered.
        delete_cache: Whether to delete the cache after segmentation.
        recover: If `True`, recover the cache from a failed segmentation, and continue.
        key_added: Name of the shapes element to be added to `sdata`.
        patch_index: Index of the patch to segment (we do not recommend to set this argument). By default, segment all patches.
    """
    _check_transcript_patches(sdata, with_prior=True)

    if config is None or not len(config):
        config = _get_default_config(sdata, sdata.shapes[SopaKeys.TRANSCRIPTS_PATCHES])
    elif isinstance(config, str):
        with open(config) as f:
            config = json.load(f)

    if "gene_column" not in config:
        config["gene_column"] = str(get_feature_key(sdata, raise_error=True))

    config["prior_name"] = SopaKeys.SOPA_PRIOR

    if patch_index is not None:
        patch_dir = get_transcripts_patches_dirs(sdata, patch_index)
        comseg_patch(patch_dir, config, recover)
        return

    patches_dirs = get_transcripts_patches_dirs(sdata)

    _functions = [partial(comseg_patch, patch_dir, config, recover) for patch_dir in patches_dirs]
    settings._run_with_backend(_functions)

    resolve(sdata, patches_dirs, config["gene_column"], min_area=min_area, key_added=key_added)

    set_boundaries_attrs(sdata, key_added)

    if delete_cache:
        delete_transcripts_patches_dirs(sdata)


def comseg_patch(patch_dir: Path, config: dict, recover: bool = False):
    import json

    try:
        import comseg
        from comseg import dataset as ds
        from comseg import dictionary
    except ModuleNotFoundError:
        raise ModuleNotFoundError("Install comseg (`pip install comseg`) for this method to work")

    assert comseg.__version__ >= "1.8.2", "comseg version should be >= 1.8.2"

    if (
        recover
        and (patch_dir / "segmentation_counts.h5ad").exists()
        and (patch_dir / "segmentation_polygons.json").exists()
    ):
        return

    if "disable_tqdm" not in config:
        config["disable_tqdm"] = True

    warnings.filterwarnings("ignore", message="param_sctransform is none")
    warnings.filterwarnings("ignore", message="Series.__getitem__")

    with HiddenPrints():
        dataset = ds.ComSegDataset(
            path_dataset_folder=patch_dir,
            dict_scale=config["dict_scale"],
            mean_cell_diameter=config["mean_cell_diameter"],
            gene_column=config["gene_column"],
            image_csv_files=["transcripts.csv"],
            centroid_csv_files=["centroids.csv"],
            path_cell_centroid=patch_dir,
            min_nb_rna_patch=config.get("min_nb_rna_patch", 0),
            prior_name=config["prior_name"],
            disable_tqdm=config["disable_tqdm"],
        )

        dataset.compute_edge_weight(config=config)

        Comsegdict = dictionary.ComSegDict(
            dataset=dataset,
            mean_cell_diameter=config["mean_cell_diameter"],
            disable_tqdm=config["disable_tqdm"],
        )

        Comsegdict.run_all(config=config)

        assert config.get("return_polygon", True), "Only return_polygon=True is supported in sopa"

        anndata_comseg, json_dict = Comsegdict.anndata_from_comseg_result(config=config)

        anndata_comseg.write_h5ad(patch_dir / "segmentation_counts.h5ad")
        with open(patch_dir / "segmentation_polygons.json", "w") as f:
            json.dump(json_dict["transcripts"], f)


def _get_default_config(sdata: SpatialData, patches: gpd.GeoDataFrame) -> dict:
    prior_shapes_key = patches[SopaKeys.PRIOR_SHAPES_KEY].iloc[0]
    points_key = patches[SopaKeys.POINTS_KEY].iloc[0]

    cells_area = to_intrinsic(sdata, prior_shapes_key, points_key).area
    mean_cell_diameter = 2 * np.sqrt(cells_area / np.pi).mean()

    config = {
        "dict_scale": {"x": 1, "y": 1, "z": 1},  # spot coordinates already in µm
        "mean_cell_diameter": mean_cell_diameter,
        "max_cell_radius": mean_cell_diameter * 1.75,
        "norm_vector": False,
        "alpha": 0.5,  # alpha value to compute the polygon https://pypi.org/project/alphashape/
        "allow_disconnected_polygon": False,
        "min_rna_per_cell": 20,  # minimal number of RNAs for a cell to be taken into account
        "gene_column": str(get_feature_key(sdata, raise_error=True)),
    }

    log.info(f"The Comseg config was not provided, using the following by default:\n{config}")

    return config


class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout



import logging
import re
from math import ceil

import numpy as np
import tifffile as tf
from multiscale_spatial_image import to_multiscale
from spatialdata import SpatialData
from spatialdata.transformations import Affine, Sequence, get_transformation, set_transformation
from tqdm import tqdm
from xarray import DataArray, DataTree

from ...utils import add_spatial_element, get_spatial_image, resize_numpy, scale_dtype
from ._constants import ExplorerConstants, FileNames, image_metadata
from .utils import explorer_file_path

log = logging.getLogger(__name__)

TILE_SIZE = ExplorerConstants.TILE_SIZE


class MultiscaleImageWriter:
    photometric = "minisblack"
    compression = "jpeg2000"
    resolutionunit = "CENTIMETER"
    dtype = np.uint8

    def __init__(self, image: DataTree, tile_width: int, pixel_size: float):
        self.image = image
        self.tile_width = tile_width
        self.pixel_size = pixel_size

        self.scale_names = list(self.image.children)
        self.channel_names = list(map(str, self.image[self.scale_names[0]].c.values))
        self.channel_names = _set_colors(self.channel_names)
        self.metadata = image_metadata(self.channel_names, pixel_size)
        self.data = None

        self.lazy = True
        self.ram_threshold_gb = None

    def _n_tiles_axis(self, xarr: DataArray, axis: int) -> int:
        return ceil(xarr.shape[axis] / self.tile_width)

    def _get_tiles(self, xarr: DataArray):
        for c in range(xarr.shape[0]):
            for index_y in range(self._n_tiles_axis(xarr, 1)):
                for index_x in range(self._n_tiles_axis(xarr, 2)):
                    tile = xarr[
                        c,
                        self.tile_width * index_y : self.tile_width * (index_y + 1),
                        self.tile_width * index_x : self.tile_width * (index_x + 1),
                    ]
                    yield self._scale(tile.values)

    def _should_load_memory(self, shape: tuple[int, int, int], dtype: np.dtype):
        if not self.lazy:
            return True

        if self.ram_threshold_gb is None:
            return False

        itemsize = max(np.dtype(dtype).itemsize, np.dtype(self.dtype).itemsize)
        size = shape[0] * shape[1] * shape[2] * itemsize

        return size <= self.ram_threshold_gb * TILE_SIZE**3

    def _scale(self, array: np.ndarray):
        return scale_dtype(array, self.dtype)

    def _write_image_level(self, tif: tf.TiffWriter, scale_index: int, **kwargs):
        xarr: DataArray = next(iter(self.image[self.scale_names[scale_index]].values()))
        resolution = 1e4 * 2**scale_index / self.pixel_size

        if not self._should_load_memory(xarr.shape, xarr.dtype):
            n_tiles = xarr.shape[0] * self._n_tiles_axis(xarr, 1) * self._n_tiles_axis(xarr, 2)
            data = self._get_tiles(xarr)
            data = iter(tqdm(data, total=n_tiles, desc="Writing tiles"))
        else:
            if self.data is not None:
                self.data = resize_numpy(self.data, 2, xarr.dims, xarr.shape)
            else:
                log.info(f"   (Loading image of shape {xarr.shape}) in memory")
                self.data = self._scale(xarr.values)

            data = self.data

        log.info(f"   > Image of shape {xarr.shape}")
        tif.write(
            data,
            tile=(self.tile_width, self.tile_width),
            resolution=(resolution, resolution),
            metadata=self.metadata,
            shape=xarr.shape,
            dtype=self.dtype,
            photometric=self.photometric,
            compression=self.compression,
            resolutionunit=self.resolutionunit,
            **kwargs,
        )

    def __len__(self):
        return len(self.scale_names)

    def procedure(self):
        if not self.lazy:
            return "in-memory (consider lazy procedure if it crashes because of RAM)"
        if self.ram_threshold_gb is None:
            return "lazy (slower but low RAM usage)"
        return "semi-lazy (load in memory when possible)"

    def write(self, path, lazy=True, ram_threshold_gb=None):
        self.lazy = lazy
        self.ram_threshold_gb = ram_threshold_gb

        log.info(f"Writing multiscale image with procedure={self.procedure()}")

        with tf.TiffWriter(path, bigtiff=True) as tif:
            self._write_image_level(tif, 0, subifds=len(self) - 1)

            for i in range(1, len(self)):
                self._write_image_level(tif, i, subfiletype=1)


def _to_color(channel_name: str, is_wavelength: bool, colors_iterator: list):
    if is_wavelength:
        return channel_name
    if channel_name in ExplorerConstants.KNOWN_CHANNELS:
        return f"{channel_name} (color={ExplorerConstants.KNOWN_CHANNELS[channel_name]})"
    return f"{channel_name} (color={colors_iterator.pop()})"


def _set_colors(channel_names: list[str]) -> list[str]:
    """
    Trick to provide a color to all channels on the Xenium Explorer.

    But some channels colors are set to white by default. This functions allows to color these
    channels with an available wavelength color (e.g., `550`).
    """
    existing_wavelength = [bool(re.search(r"(?<![0-9])[0-9]{3}(?![0-9])", c)) for c in channel_names]
    valid_colors = [c for c in ExplorerConstants.COLORS if c != ExplorerConstants.NUCLEUS_COLOR]
    n_missing = sum(
        not is_wavelength and c not in ExplorerConstants.KNOWN_CHANNELS
        for c, is_wavelength in zip(channel_names, existing_wavelength)
    )
    colors_iterator: list = np.repeat(valid_colors, ceil(n_missing / len(valid_colors))).tolist()

    return [
        _to_color(c, is_wavelength, colors_iterator) for c, is_wavelength in zip(channel_names, existing_wavelength)
    ]


def _to_xenium_explorer_multiscale(image: DataArray | DataTree, n_subscales: int) -> DataTree:
    if isinstance(image, DataTree):
        shapes = np.array([next(iter(data_tree.values())).shape[1:] for data_tree in image.values()])
        if len(shapes) == n_subscales + 1 and (shapes[:-1] // shapes[1:] == 2).all():
            return image

        image = DataArray(next(iter(image["scale0"].values())))

    return to_multiscale(image, [2] * n_subscales, chunks=(1, TILE_SIZE, TILE_SIZE))


def write_image(
    path: str,
    image: DataTree | DataArray | np.ndarray,
    lazy: bool = True,
    tile_width: int = TILE_SIZE,
    n_subscales: int = 5,
    pixel_size: float = 0.2125,
    ram_threshold_gb: int | None = 4,
    is_dir: bool = True,
):
    """Convert an image into a `morphology.ome.tif` file that can be read by the Xenium Explorer

    Args:
        path: Path to the Xenium Explorer directory where the image will be written
        image: Image of shape `(C, Y, X)`
        lazy: If `True`, the image will not be read in-memory (except if the image size is below `ram_threshold_gb`). If `False`, all the images levels are always loaded in-memory.
        tile_width: Xenium tile width (do not update).
        n_subscales: Number of sub-scales in the pyramidal image.
        pixel_size: Xenium pixel size (do not update).
        ram_threshold_gb: If an image (of any level of the pyramid) is below this threshold, it will be loaded in-memory.
        is_dir: If `False`, then `path` is a path to a single file, not to the Xenium Explorer directory.
    """
    path = explorer_file_path(path, FileNames.IMAGE, is_dir)

    if isinstance(image, np.ndarray):
        assert len(image.shape) == 3, "Can only write channels with shape (C,Y,X)"
        log.info(f"Converting image of shape {image.shape} into a DataArray (with dims: C,Y,X)")
        image = DataArray(image, dims=["c", "y", "x"], name="image")

    image = _to_xenium_explorer_multiscale(image, n_subscales)

    image_writer = MultiscaleImageWriter(image, pixel_size=pixel_size, tile_width=tile_width)
    image_writer.write(path, lazy=lazy, ram_threshold_gb=ram_threshold_gb)


def align(
    sdata: SpatialData,
    image: DataArray,
    transformation_matrix_path: str,
    key_added: str | None = None,
    image_key: str | None = None,
    overwrite: bool = False,
):
    """Add an image to the `SpatialData` object after alignment with the Xenium Explorer.

    Args:
        sdata: A `SpatialData` object
        image: A `DataArray` object. Note that `image.name` is used as the key for the aligned image.
        transformation_matrix_path: Path to the `.csv` transformation matrix exported from the Xenium Explorer
        key_added: Optional name to add to the new image. If `None`, will use `image.name`.
        image_key: Optional name of the image on which it has been aligned. Required if multiple images in the `SpatialData` object.
        overwrite: Whether to overwrite the image, if already existing.
    """
    key_added = key_added or image.name

    assert key_added is not None, "The image has no name, use the `key_added` argument to provide one"
    assert key_added not in sdata, f"Image '{key_added}' already exists in the `SpatialData` object"

    to_pixel = Sequence([
        Affine(
            np.genfromtxt(transformation_matrix_path, delimiter=","),
            input_axes=("x", "y"),
            output_axes=("x", "y"),
        )
    ])

    default_image = get_spatial_image(sdata, image_key)

    original_transformations = get_transformation(default_image, get_all=True)
    transformations = {cs: to_pixel.compose_with(t) for cs, t in original_transformations.items()}

    set_transformation(image, transformations, set_all=True)

    add_spatial_element(sdata, key_added, image, overwrite=overwrite)
    log.info(f"Added image '{key_added}' (aligned with the Xenium Explorer)")



import ast

import typer

from .utils import SDATA_HELPER

app_annotate = typer.Typer()


@app_annotate.command()
def fluorescence(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    marker_cell_dict: str = typer.Option(callback=ast.literal_eval),
    cell_type_key: str = typer.Option("cell_type", help="Key added in `adata.obs` corresponding to the cell type"),
):
    """Simple annotation based on fluorescence, where each provided channel corresponds to one cell type.

    For each cell, one z-score statistic is computed and the population with the highest z-score is attributed.
    """
    from sopa.constants import SopaKeys
    from sopa.io.standardize import read_zarr_standardized
    from sopa.utils import higher_z_score

    sdata = read_zarr_standardized(sdata_path)

    assert SopaKeys.TABLE in sdata.tables, f"No '{SopaKeys.TABLE}' found in sdata.tables"

    higher_z_score(sdata.tables[SopaKeys.TABLE], marker_cell_dict, cell_type_key)

    if sdata.is_backed():
        sdata.delete_element_from_disk(SopaKeys.TABLE)
        sdata.write_element(SopaKeys.TABLE)


@app_annotate.command()
def tangram(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    sc_reference_path: str = typer.Option(help="Path to the scRNAseq annotated reference, as a `.h5ad` file"),
    cell_type_key: str = typer.Option(help="Key of `adata_ref.obs` containing the cell-types"),
    reference_preprocessing: str = typer.Option(
        None,
        help="Preprocessing method applied to the reference. Either None (raw counts), or `normalized` (sc.pp.normalize_total) or `log1p` (sc.pp.normalize_total and sc.pp.log1p)",
    ),
    bag_size: int = typer.Option(
        10_000,
        help="Number of cells in each bag of the spatial table. Low values will decrease the memory usage",
    ),
    max_obs_reference: int = typer.Option(
        10_000,
        help="Maximum samples to be considered in the reference for tangram. Low values will decrease the memory usage",
    ),
):
    """Tangram segmentation (i.e., uses an annotated scRNAseq reference to transfer cell-types)"""
    import anndata

    from sopa.constants import SopaKeys
    from sopa.io.standardize import read_zarr_standardized
    from sopa.utils import tangram_annotate

    sdata = read_zarr_standardized(sdata_path)
    adata_sc = anndata.io.read_h5ad(sc_reference_path)

    tangram_annotate(
        sdata,
        adata_sc,
        cell_type_key,
        reference_preprocessing=reference_preprocessing,
        bag_size=bag_size,
        max_obs_reference=max_obs_reference,
    )
    if sdata.is_backed():
        sdata.delete_element_from_disk(SopaKeys.TABLE)
        sdata.write_element(SopaKeys.TABLE, overwrite=True)
